#
#	FILE:	 Grid.py
#	AUTHOR:  Marcus Turner
#	PURPOSE: Global map script - based on Continents.pl
#-----------------------------------------------------------------------------
#	Copyright (c) 2004, 2005 Firaxis Games, Inc. All rights reserved.
#-----------------------------------------------------------------------------
#

from CvPythonExtensions import *
import CvUtil
import CvMapGeneratorUtil
from CvMapGeneratorUtil import FractalWorld
from CvMapGeneratorUtil import TerrainGenerator
from CvMapGeneratorUtil import FeatureGenerator

def getDescription():
	return "TXT_KEY_MAP_SCRIPT_GRID_DESCR_V005"
	
def isAdvancedMap():
	"This map should show up in simple mode"
	return 0

# Custom options
# [TODO] Generate a resource file for the various entries

# - Border land type
# - Resource allocation
# - Linked tiles
# - Allow rivers
# - Allow goodie huts

def getNumCustomMapOptions():
	return 5

def getCustomMapOptionName(argsList):
	"""
	Returns name of specified option
	argsList[0] is Option ID (int)
	"""
	[iOption] = argsList
	option_names = {
		0:	"TXT_KEY_MAP_SCRIPT_BORDER_TYPE",
		1:	"TXT_KEY_MAP_SCRIPT_RESOURCE_ALLOCATION",
		2:	"TXT_KEY_MAP_SCRIPT_LINK_TILES",
		3:	"TXT_KEY_MAP_SCRIPT_ALLOW_RIVERS",
		4:	"TXT_KEY_MAP_SCRIPT_ALLOW_GOODIES"
		}
	translated_text = unicode(CyTranslator().getText(option_names[iOption], ()))
	return translated_text
	
def getNumCustomMapOptionValues(argsList):
	[iOption] = argsList
	option_values = {
		0:	4,
		1:	3,
		2:	2,
		3:	2,
		4: 	2
		}
	return option_values[iOption]
	
def getCustomMapOptionDescAt(argsList):
	[iOption, iSelection] = argsList
	selection_names = {
		0:	{
			0: "TXT_KEY_MAP_SCRIPT_OCEAN",
			1: "TXT_KEY_MAP_SCRIPT_PLAIN",
			2: "TXT_KEY_MAP_SCRIPT_HILL",
			3: "TXT_KEY_MAP_SCRIPT_PEAK"
			},
		1:	{
			0: "TXT_KEY_MAP_SCRIPT_HUGE",
			1: "TXT_KEY_MAP_SCRIPT_NORMAL",
			2: "TXT_KEY_MAP_SCRIPT_NONE"
			},
		2:	{
			0: "TXT_KEY_MAP_SCRIPT_YES",
			1: "TXT_KEY_MAP_SCRIPT_NO"
			},
		3:	{
			0: "TXT_KEY_MAP_SCRIPT_YES",
			1: "TXT_KEY_MAP_SCRIPT_NO"
			},
		4:	{
			0: "TXT_KEY_MAP_SCRIPT_YES",
			1: "TXT_KEY_MAP_SCRIPT_NO"
			}
		}
	translated_text = unicode(CyTranslator().getText(selection_names[iOption][iSelection], ()))
	return translated_text
	
def getCustomMapOptionDefault(argsList):
	[iOption] = argsList
	option_defaults = {
		0:	0,
		1:	0,
		2:	0,
		3:	0,
		4:	0
		}
	return option_defaults[iOption]
	

# Main entry points

def generatePlotTypes():
	NiTextOut("Setting Plot Types (Python Grid) ...")
	fractal_world = GridFractalWorld()
	# fractal_world.initFractal(polar = True)
	return fractal_world.generatePlotTypes()

def generateTerrainTypes():
	NiTextOut("Generating Terrain (Python Grid) ...")
	terraingen = GridTerrainGenerator()
	terrainTypes = terraingen.generateTerrain()
	return terrainTypes

def addFeatures():
	NiTextOut("Adding Features (Python Grid) ...")
	featuregen = GridFeatureGenerator()
	featuregen.addFeatures()
	return 0
		
		
# Pre-generation setting of various size parameters for this world

def beforeGeneration():

	# Calculate X/Y offsets for the map so that it is centered and we get a
	# regular number of plots.  Use these values to calculate the number of
	# 3x3 fat cross plots we can allocate
	
	global gridStartPositions
	gridStartPositions = []		
	
	# Allow for an increased border between 3x3 squares
	global gridOffset
	gridOffset = 0

	# Get the number of 6x6 tiles we can add to the map
	iH = CyMap().getGridHeight()
	iW = CyMap().getGridWidth()

	modX, remX = divmod(iW, 6)
	modY, remY = divmod(iH, 6)

	# Calculate the number of 3x3 plot of 6x6 we can fit in the map
	gridXPlots, rem3by3X = divmod(modX, 3)
	gridYPlots, rem3by3Y = divmod(modY, 3)

	# If we have no remainder then the 3x3 grids fit exactly in the map, so
	# disable the offset
	if (remX == 0) or (remX == 0):
		gridOffset = 0
	
	# Calculate the centered x/y max for the grid placement if there is not an exact fit
	maxX = iW - 10 - (rem3by3X * 3) + ((gridOffset * (gridXPlots - 1)) / 2)
	if (remX != 0):
		maxX = maxX - ((remX - 1) / 2)

	maxY = iH - 10 - (rem3by3Y * 3) + ((gridOffset * (gridYPlots - 1)) / 2)
	if (remY != 0):
		maxY = maxY - ((remY - 1) / 2)
	
	# Debugging
	print "Map dimensions; W", iW, "H", iH, "maxX", maxX, "maxY", maxY

	# Now that we know the max X/Y, just add starting plots from the top right decrementing
	# backwards until we can fit no more

	# We can also calculate the maximum number of civilisations we can support
	global maxCivs
	maxCivs = 0

	y = maxY
	while (y >= 9):
		x = maxX
		while (x >= 9):
			# Append each center point to the starting position set
			gridStartPositions.append( (x, y) )
			
			print "Calculated start position:", (x, y)

			x = (x - 18) - gridOffset
			maxCivs = maxCivs + 1
			
	
		# Decrement the y and redo the next row
		y = (y - 18) - gridOffset
		
	# Create a series of grid masks
	# These define offsets around the center of a 5x5 tile, so that we can apply changes later
	
	# Default layout is
	# C - Corner  c - InsideCorner  m - Inside  X - center  O - Outside
	#
	# OOOOOO
	# CcmcCO
	# cmmmCO
	# mmXmmO
	# cmmmcO
	# CcmcCO
	
	global tileOffsetCorners 
	tileOffsetCorners = [(-2, -2), (-2, 2), (2, 2), (2, -2)]
	
	global tileOffsetInsideCorners
	tileOffsetInsideCorners = [(-2, -1), (-2, 1), (-1, -2), (-1, 2),
	                           (2, -1), (2, 1), (1, -2), (1, 2)]
	
	global tileOffsetInside
	tileOffsetInside = [(-2, 0), (-1, -1), (-1, 0), (-1, 1),
	                     (0, -2), (0, -1), (0, 1), (0, 2),
	                     (2, 0), (1, -1), (1, 0), (1, 1)]

	global tileOffsetOutside
	tileOffsetOutside = [(-2, 3), (-1, 3), (0, 3), (1, 3), (2, 3), (3, 3),
	                     (3, 2), (3, 1), (3, 0), (3, -1), (3, -2)]

	global tileOffsetSeaLink
	tileOffsetSeaLink = [(0, -1), (0, -2), (1, -2)]

	                        
	# Create a grid mask for the internal linking of the 3x3 grids
	global tileOffsetLinks
	tileOffsetLinks = [(-3, 6), (-3, 0), (-3, -6), (0, -3), (0, 3), (3, -6), (3, 0), (3, 6)]
	
	# Given the initial map option to choose the fill type between tiles,
	# calculate and store the correct plot type
	plotTypes = [ PlotTypes.PLOT_OCEAN, PlotTypes.PLOT_LAND, PlotTypes.PLOT_HILLS, PlotTypes.PLOT_PEAK ]

	global defaultPlotType
	defaultPlotType = plotTypes[CyMap().getCustomMapOption(0)]
	
	# We should now have a list of valid starting positions for this size of map,
	# check to see if we have too many players

	# Success flag. Set to false if number of players falls outside legal range.
	global bSuccessFlag
	
	iPlayers = CyGlobalContext().getGame().countCivPlayersEverAlive()
	if (iPlayers < 2 or iPlayers > maxCivs):
		bSuccessFlag = False
	else:
		bSuccessFlag = True

# Allocate starting plots

def assignStartingPlots():

	gc = CyGlobalContext()

	# If we have more players than 3x3 segments default to random placement
	global bSuccessFlag
	if bSuccessFlag == False:
		return CyPythonMgr().allowDefaultImpl()

	# Grab the player list
	player_list = []
	for plrCheckLoop in range(18):
		if CyGlobalContext().getPlayer(plrCheckLoop).isEverAlive():
			player_list.append(plrCheckLoop)

	# Assign each player one of the starting positions, using a copy of the
	# start positions so that we can pop them off as needed
	startPositions = gridStartPositions
	
	for playerID in player_list:
		player = gc.getPlayer(playerID)
		
		midX, midY = startPositions.pop()
		sPlot = CyMap().sPlot(midX, midY)

		player.setStartingPlot(sPlot, true)	


# Disabled to be added later if wanted

def addRivers():
	# Add rivers if required
	if (CyMap().getCustomMapOption(3) == 0):
		return CyPythonMgr().allowDefaultImpl()
	else:
		return 0

def addLakes():
	NiTextOut("Disabled addLakes (Python Grid) ...")
	return 0

def addBonuses():
	# Bonuses are added as part of addResource if set to huge, otherwise
	# return none or the default
	if (CyMap().getCustomMapOption(1) == 1):
		# Normal random resource generation
		return CyPythonMgr().allowDefaultImpl()
	else:
		return 0

def addGoodies():
	# Add goodies if required
	if (CyMap().getCustomMapOption(4) == 0):
		return CyPythonMgr().allowDefaultImpl()
	else:
		return 0


# Not required as this is a purely artificial map

def normalizeStartingPlotLocations():
	return None

def normalizeAddRiver():
	return None

def normalizeRemovePeaks():
	return None

def normalizeAddLakes():
	return None

def normalizeRemoveBadFeatures():
	return None

def normalizeRemoveBadTerrain():
	return None

def normalizeAddFoodBonuses():
	return None

def normalizeAddGoodTerrain():
	return None

def normalizeAddExtras():
	return None


# Subclass to customize plot generation
class GridFractalWorld(CvMapGeneratorUtil.FractalWorld):
	
	def generatePlotTypes(self, water_percent=25, shift_plot_types=True, grain_amount=3):

		print "GridFractalWorld.GeneratePlotTypes"
		
		# self.plotTypes[i] = PlotTypes.PLOT_OCEAN
		# self.plotTypes[i] = PlotTypes.PLOT_LAND
		# self.plotTypes[i] = PlotTypes.PLOT_HILLS
		# self.plotTypes[i] = PlotTypes.PLOT_PEAK

		# Create an empty grid, in this case of ocean
		# If you want a plain grid, change PLOT_OCEAN to PLOT_LAND and
		# then immediately exit the function after self.plotTypes
		global defaultPlotType
		
		self.plotTypes = []
		self.plotTypes = [defaultPlotType] * (self.iNumPlotsX*self.iNumPlotsY)

		# Generate each of the 3x3 grids around the starting positions
		global gridStartPositions
		for midX, midY in gridStartPositions:
			self.generateTile3x3(midX, midY)

		# Link the different grids together if required
		if (CyMap().getCustomMapOption(2) == 0):
			
			# There must be a nicer way of doing this, but as this is an initialisation
			# and not every turn check every starting plot against every other and link
			# to the right or up if the plots are next to each other
			
			for startX, startY in gridStartPositions:
			 	for endX, endY in gridStartPositions:
			 		self.checkGridLink(startX, endX, startY, endY)

		# Finished... don't shift the plots though as we should be centered already
		# if shift_plot_types:
		# 	self.shiftPlotTypes()

		return self.plotTypes

	def checkGridLink(self, startX, endX, startY, endY):

		global gridOffset

		# If we have the correct minimum distance between the grids
		# add the link horizontally
		if ((endX - startX) == (18 + gridOffset)):
			for x in range(startX + 9, endX - 8):
	 			self.generateTile(x, startY, [(0,0)], PlotTypes.PLOT_HILLS)
	
		# If we have the correct minimum distance between the grids
		# add the link vertically
		if ((endY - startY) == (18 + gridOffset)):
			for y in range(startY + 9, endY - 8):
	 			self.generateTile(startX, y, [(0,0)], PlotTypes.PLOT_HILLS)

	def generateTile3x3(self, midX, midY):
	 	
	 	# Generate a 3x3 plot around the supplied point
		global tileOffsetCorners 
		global tileOffsetInsideCorners
		global tileOffsetInside
		global tileOffsetOutside
		global tileOffsetSeaLink
		global tileOffsetLinks
		global defaultPlotType
		
		# Debugging
		print "GenerateTile3x3: ", midX, midY
	 	
	 	# Use the tile offsets to fill in the details
	 	for x in range(midX - 6, midX + 7, 6):
	 		for y in range(midY - 6, midY + 7, 6):
	 		
	 			self.generateTile(x, y, [(0,0)], PlotTypes.PLOT_HILLS)

	 			self.generateTile(x, y, tileOffsetCorners, PlotTypes.PLOT_LAND)
	 			self.generateTile(x, y, tileOffsetInsideCorners, PlotTypes.PLOT_HILLS)
	 			self.generateTile(x, y, tileOffsetInside, PlotTypes.PLOT_LAND)
	 			self.generateTile(x, y, tileOffsetOutside, defaultPlotType)

				# By default, each of the center crosses has no access to the sea, and
				# so there has to be a choice as to whether you build near the coast
				# or not.  However, if you want to bypass this and link each middle square
				# to the sea uncomment the following
				
				# self.generateTile(x, y, tileOffsetSeaLink, PlotTypes.PLOT_OCEAN)
				
		# Link the 3x3 entries as appropriate
		self.generateTile(midX, midY, tileOffsetLinks, PlotTypes.PLOT_HILLS)


	def generateTile(self, x, y, tileOffsets, plotType):

		# Debugging
		print "GenerateTile: ", x, y, tileOffsets, plotType

		# For each offset, apply the tile plot type
		for offsetX, offsetY in tileOffsets:
			i = (y + offsetY) * self.iNumPlotsX + (x + offsetX)
			self.plotTypes[i] = plotType


# Subclass to customize terrain generation
class GridTerrainGenerator(CvMapGeneratorUtil.TerrainGenerator):

	def generateTerrainAtPlot(self,iX,iY):
		if (self.map.plot(iX, iY).isWater()):
			return self.map.plot(iX, iY).getTerrainType()
			
		# Disable terrain type generation for everything else
		return 0


# Subclass to customize feature generation
class GridFeatureGenerator(CvMapGeneratorUtil.FeatureGenerator):

	def addFeatures(self):
		
		# If normal bonus allocation is enabled, just return here

		# Huge resources will provide:
		# - Metal under each potential starting square
		# - Coal/Iron on each hill
		# - Wheat/Corn on each plain
		# - One of every other resource in the corners
		
		self.bonusWheat = self.gc.getInfoTypeForString("BONUS_WHEAT")
		self.bonusCorn = self.gc.getInfoTypeForString("BONUS_CORN")
		
		self.bonusGold = self.gc.getInfoTypeForString("BONUS_GOLD")
		self.bonusIron = self.gc.getInfoTypeForString("BONUS_IRON")
		self.bonusCopper = self.gc.getInfoTypeForString("BONUS_COPPER")
		self.bonusStone = self.gc.getInfoTypeForString("BONUS_STONE")
		self.bonusMarble = self.gc.getInfoTypeForString("BONUS_MARBLE")
		
		self.bonusFish = self.gc.getInfoTypeForString("BONUS_FISH")
		self.bonusWhale = self.gc.getInfoTypeForString("BONUS_WHALE")
		self.bonusCrab = self.gc.getInfoTypeForString("BONUS_CRAB")
		self.bonusClam = self.gc.getInfoTypeForString("BONUS_CLAM")
		
		self.bonusMovies = self.gc.getInfoTypeForString("BONUS_MOVIES")
		self.bonusMusic = self.gc.getInfoTypeForString("BONUS_MUSIC")
		self.bonusDrama = self.gc.getInfoTypeForString("BONUS_DRAMA")

		if (CyMap().getCustomMapOption(1) == 0):
			# Huge resources
			global gridStartPositions
			for midX, midY in gridStartPositions:
				self.addHugeResources(midX, midY)
		else:
			# None
			return 0

	def addHugeResources(self, midX, midY):

	 	# Generate a 3x3 resource plot around the supplied point
		global tileOffsetCorners 
		global tileOffsetInsideCorners
		global tileOffsetInside
		global tileOffsetOutside
		global tileOffsetLinks

		# Create a list of the bonustypes so that we can remove those that are used
		# as we apply them...
		
		# [TODO] I can't figure out at the moment how to get the bonus valid checking
		#        for a terrain feature working, so remove all the ocean type products
		#        from the list... horrible...
		bonusList = []
		for bonusType in range(self.gc.getNumBonusInfos()):
			if ((bonusType != self.bonusFish) and
			    (bonusType != self.bonusWhale) and
			    (bonusType != self.bonusCrab) and
			    (bonusType != self.bonusClam) and
			    (bonusType != self.bonusMovies) and
			    (bonusType != self.bonusMusic) and
			    (bonusType != self.bonusDrama)):
				bonusList.append(bonusType)

		# Randomise the bonus distribution
		# Ideally, if this were to be used to guarantee identical map layouts, we would
		# want to be much more clever...
		# [TODO] Implement the bonus list randomisation


		# Iron and copper on the hills, extend or change the list
		# if you want other materials
		metalList = [ self.bonusIron, self.bonusCopper ]
		
		# Stone and marble in the middle
		stoneList = [ self.bonusStone, self.bonusMarble ]

		# Sea resources
		seaList = [ self.bonusFish, self.bonusCrab, self.bonusClam, self.bonusWhale ]

	 	# Use the tile offsets to fill in the details
	 	
	 	# [TODO] For some reason the resources end up offset on the X axis by +1, so for
	 	#        the moment add an artificial shift left of 1
	 	# for x in range(midX - 6, midX + 7, 6):
	 	for x in range(midX - 7, midX + 7, 6):
	 		for y in range(midY - 6, midY + 7, 6):
	 			
	 			# And a metal in the center
				if (len(stoneList) > 0):
					stoneItem = stoneList.pop(0)
					self.addLandBonus(x, y, [(0, 0)], stoneItem)
					stoneList.append(stoneItem)
	 			
	 			# Forest plus resources in the corners
	 			self.addResource(x, y, tileOffsetCorners, self.featureForest)

	 			# One bonus feature at each corner
				for cornerXY in tileOffsetCorners:
		 			if len(bonusList) > 0:
		 				bonusItem = bonusList.pop(0)
						self.addLandBonus(x, y, [cornerXY], bonusItem)
						bonusList.append(bonusItem)
	 			
	 			# If going for a major resource boost, use these as well
	 			
	 			# Wheat or corn in the middle cross
	 			self.addLandBonus(x, y, tileOffsetInside, self.bonusWheat)
	 			
	 			# Fish or similar in the coastal area, if it has been created
				for cornerXY in tileOffsetSeaLink:
		 			if len(seaList) > 0:
		 				seaItem = seaList.pop(0)
						self.addWaterBonus(x, y, [cornerXY], seaItem)
						seaList.append(seaItem)

	 			# Metals on the hills
				for cornerXY in tileOffsetInsideCorners:
					if (len(metalList) > 0):
						metalItem = metalList.pop(0)
						self.addLandBonus(x, y, [cornerXY], metalItem)
						metalList.append(metalItem)

	def addResource(self, x, y, tileOffsets, featureType):
		# Debugging
		print "AddResource: ", x, y, tileOffsets, featureType

		# For each offset, apply the feature
		for offsetX, offsetY in tileOffsets:
			pPlot = self.map.sPlot(x + offsetX + 1, y + offsetY)
			pPlot.setFeatureType(featureType, -1)
		
	def addLandBonus(self, x, y, tileOffsets, bonusType):
		# Debugging
		print "addLandBonus: ", x, y, tileOffsets, bonusType

		# For each offset, apply the bonus
		for offsetX, offsetY in tileOffsets:
			pPlot = self.map.sPlot(x + offsetX + 1, y + offsetY)
			# [TODO] Make this automatic rather than hardcoded, so appropriate
			#        resources only ever go on the right tile types
			if (not pPlot.isWater()):
				pPlot.setBonusType(bonusType)

	def addWaterBonus(self, x, y, tileOffsets, bonusType):
		# Debugging
		print "addWaterBonus: ", x, y, tileOffsets, bonusType

		# For each offset, apply the bonus
		for offsetX, offsetY in tileOffsets:
			pPlot = self.map.sPlot(x + offsetX + 1, y + offsetY)
			# [TODO] Make this automatic rather than hardcoded, so appropriate
			#        resources only ever go on the right tile types
			if (pPlot.isWater()):
				pPlot.setBonusType(bonusType)

	def addFeaturesAtPlot(self, iX, iY):
		# Disable feature generation
		return 0

