# FILE: LandMasses.py
# PURPOSE: Provide an alternative continents style map script.
# VERSION 2.90

# Import the required files.

import operator
from CvPythonExtensions import *
import CvUtil
import CvMapGeneratorUtil
from CvMapGeneratorUtil import FractalWorld
from CvMapGeneratorUtil import TerrainGenerator
from CvMapGeneratorUtil import FeatureGenerator
import pickle
import math

#--------------------------------------------------------------------------------------------------------------------------#
# CUSTOMISING RESOURCE PLACEMENT
#--------------------------------------------------------------------------------------------------------------------------#

# Note the UNMODDED value of each resource.
#  0 - Aluminium
#  1 - Coal
#  2 - Copper
#  3 - Horse
#  4 - Iron
#  5 - Marble
#  6 - Oil
#  7 - Stone
#  8 - Uranium
#  9 - Banana
# 10 - Clam
# 11 - Corn
# 12 - Cow
# 13 - Crab
# 14 - Deer
# 15 - Fish
# 16 - Pig
# 17 - Rice
# 18 - Sheep
# 19 - Wheat
# 20 - Dye
# 21 - Fur
# 22 - Gems
# 23 - Gold
# 24 - Incense
# 25 - Ivory
# 26 - Silk
# 27 - Silver
# 28 - Spices
# 29 - Sugar
# 30 - Wine
# 31 - Whale

# CUSTOMISING 'RESOURCE LANDMASSES' OPTION
# Editing here will alter which resources are restricted to individual land masses by the option 'Custom' under the custom
# game setup menu option 'Resource LandMasses:'. Left unedited, this option will give the same results as 'LandMasses'.

# Isolated contains the list of resources which are restricted to a single continent.

def CustomiseResourceLandMasses():
        
        Isolated = [11, 12, 16, 17, 18, 19, 25, 26, 29, 30]
        
        return Isolated

# CUSTOMISING 'HOW MANY LAND MASSES' OPTION
# Editing here will alter how many land masses each resource is assigned to by the option "Custom" under the custom game
# setup menu option 'How Many Land Masses?'. Left unedited, this option will assign isolated resources to just one land
# mass. Resources are listed in the above order, starting with the 0th resource, aluminium, and ending with the 31st
# resource, whale.
# Values can be -2, -1, 0, or from 1 to 9, or some variant of LandMassNumber.
# -2 assigns a resource to all but one of the land masses.
# -1 assigns a resource to half of the land masses.
# 0 assigns a resource to all land masses.
# 1 to 9 assigns a resource to that many land masses, assuming that many exist.
# A number less than -2, greater than 9 or greater than the number of landmasses is treated as a 0.
# LandMassNumber is the number of different land masses generated by the game. This may be different from the number of
# land masses selected in the custom game setup menu. Setting a resource to 'LandMassNumber - 2' would place a resource
# on all land masses but two, unless there are two land masses or less, in which case the resource is placed on all land
# masses. Note that Python is case sensitive. You must place the capital letters correctly to make use of LandMassNumber.

def CustomisingHowManyLandMasses(LandMassNumber):

        Isolations = [1, 1, 1, 1, 1, 1, 1, 1, 1,          # Strategic Resources.
                      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,    # Health Resources.
                      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1] # Luxury Resources.

        return Isolations

# CUSTOMISING 'RESOURCE TILES' OPTION
# Editing here will alter where resources will be placed by the option 'Custom' under the custom game setup menu
# option 'Resource Tiles:'. Left unedited, this option will give the same results as 'LandMasses'.

# AllLand contains the resources that can be found on any land tile. A resource placed in AllLand should not be placed in
# any other list relating to land tiles.
# AllWater contains the resources that can be found on any salt water tile (within reach of a coastal city's fat cross). A
# resource placed in AllWater should not be placed in any other list relating to water tiles.
# Coast contains the resources that can be found in coastal waters.
# Ocean contains the resources that can be found in ocean waters.
# Snow lists the resources that can be found on flat snow.
# HillSnow lists the resources that can be found on snow hills.
# Tundra lists the resources that can be found on flat tundra.
# ForestTundra lists the resources that can be found on flat forested tundra.
# HillTundra lists the resources that can be found on tundra hills.
# HillForestTundra lists the resources that can be found on forested tundra hills.
# And so on.

def CustomiseResourceTiles():
        
        AllLand = [0, 1, 2, 4, 5, 6, 7, 8, 22, 23, 24, 27]
        AllWater = [6, 10, 13, 15, 31]
        Coast = []
        Ocean = []
        Snow = [21]
        HillSnow = [21]
        Tundra = [14, 18, 21, 25]
        ForestTundra = [14, 21, 25]
        HillTundra = [14, 18, 21]
        HillForestTundra = [14, 21]
        FloodPlainTundra = []
        Grass = [3, 11, 12, 16, 17, 18, 19, 20, 25, 28, 29, 30]
        ForestGrass = [9, 16, 20, 25, 26, 28, 29, 30]
        JungleGrass = [9, 16, 20, 25, 26, 28, 29, 30]
        HillGrass = [16, 18, 20, 28, 30]
        HillForestGrass = [9, 16, 20, 26, 28, 29, 30]
        HillJungleGrass = [9, 16, 20, 26, 28, 29, 30]
        FloodPlainGrass = []
        Plain = [3, 11, 12, 14, 16, 17, 18, 19, 20, 21, 25, 28, 29, 30]
        ForestPlain = [9, 14, 16, 20, 21, 25, 26, 28, 29, 30]
        HillPlain = [14, 16, 18, 20, 21, 28, 30]
        HillForestPlain = [9, 14, 16, 20, 21, 26, 28, 29, 30]
        FloodPlainPlain = []
        Desert = []
        HillDesert = []
        FloodPlainDesert = []

        return (AllLand, AllWater, Coast, Ocean, Snow, HillSnow, Tundra, ForestTundra, HillTundra, HillForestTundra,
                FloodPlainTundra, Grass, ForestGrass, JungleGrass, HillGrass, HillForestGrass, HillJungleGrass,
                FloodPlainGrass, Plain, ForestPlain, HillPlain, HillForestPlain, FloodPlainPlain, Desert, HillDesert,
                FloodPlainDesert)

# CUSTOMISING RESOURCE QUANTITIES (STRATEGIC, HEALTH, AND LUXURY) OPTIONS
# ResourceQuantities lists the number of each resource to be placed. Resources are listed in the above order, starting
# with the 0th resource, aluminium, and ending with the 31st resource, whale.
# Unedited, selecting 'custom' from any of the resource quantity options will set such resources to 'normal'.
# Values can be -4, -3, -2, -1, 0, or greater.
# 0 or greater produces that amount of the appropriate resource.
# Setting a resource to -1 will produce few of that resource as defined by the resource quantity options 'few'.
# Setting a resource to -2 will produce a normal amount of it as defined by the resource quantity options 'normal'.
# Setting a resource to -3 will produce many of that resource as defined by the resource quantity options 'many'.
# Setting a resource to -4 will produce one resource per civilization.
# If a resource is set to -5 or less it will be treated as -2.

def CustomiseResourceQuantities():
        
        ResourceQuantities = [-2, -2, -2, -2, -2, -2, -2, -2, -2,             # StrategicResources.
                              -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,     # Health Resources.
                              -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2] # Luxury resources.

        return ResourceQuantities

# CUSTOMISING 'LET RESOURCES CLUMP' OPTION
# ResourceClumps lists whether or not each resource may be clumped. Resources are listed in the above order, starting
# with the 0th resource, aluminium, and ending with the 31st resource, whale.
# Unedited, selecting 'custom' from any of the resource quantity options will set such resources to 'standard'.
# Values can be 0 (resource will not be clumped) or 1 (resource may be clumped).
# If a resource is set to any other value it will be treated as 1.

def CustomiseResourceClumps():

        ResourceClumps = [0, 0, 0, 0, 0, 0, 0, 0, 0,          # Strategic Resources.
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,    # Health Resources.
                          1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1] # Luxury Resources.

        return ResourceClumps

#--------------------------------------------------------------------------------------------------------------------------#
# Predefined variables.
#--------------------------------------------------------------------------------------------------------------------------#

width = 0
height = 0
Area = 0
fillType = PlotTypes.PLOT_OCEAN
cachedMenuchoices = []
NewplotTyeps = []
UpdatedOwnerLandMasses = []
LandMassTileCount = []
Details = 0
AssignedStartTiles = []

cardinalDirs = [(-1,0),(0,-1),(0,1),(1,0)]

# Relative to 'c', cardinalDirs covers the tiles 'Y' but does not cover the tiles '.':

# .....
# ..Y..
# .YcY.
# ..Y..
# .....

nearDirs = [(-1,-1),(-1,0),(-1,1),(0,-1),(0,1),(1,-1),(1,0),(1,1)]

# Relative to 'c', nearDirs covers the tiles 'Y' but does not cover the tiles '.':

# .....
# .YYY.
# .YcY.
# .YYY.
# .....

PlayerStartCross = [(-1,-1),(-1,0),(-1,1),(0,-1),(0,1),(1,-1),(1,0),(1,1),
                    (-2,-1),(-2,0),(-2,1),(2,-1),(2,0),(2,1),(-1,2),(0,2),(1,2),(-1,-2),(0,-2),(1,-2)]

# Relative to 'c', PlayerStartCross covers the tiles 'Y' but does not cover the tiles '.':

# .......
# ..YYY..
# .YYYYY.
# .YYcYY.
# .YYYYY.
# ..YYY..
# .......

PreventResourceClumping = [(-1,4),(0,4),(1,4),(4,1),(4,0),(4,-1),(-1,-4),(0,-4),(1,-4),(-4,1),(-4,0),(-4,-1),
                           (2,3),(1,3),(0,3),(-1,3),(-2,3),(-3,2),(-3,1),(-3,0),(-3,-1),(-3,-2),(-2,-3),(-1,-3),(0,-3),
                           (1,-3),(2,-3),(3,-2),(3,-1),(3,0),(3,1),(3,2),
                           (2,2),(1,2),(0,2),(-1,2),(-2,2),(-2,1),(-2,0),(-2,-1),(-2,-2),(-1,-2),(0,-2),(1,-2),(2,-2),
                           (2,-1),(2,0),(2,1),
                           (1,1),(0,1),(-1,1),(-1,0),(-1,-1),(0,-1),(1,-1),(1,0)]

# Relative to 'r', PreventResourceClumping covers the tiles 'Y' but does not cover the tiles '.':

# ...........
# ....YYY....
# ...YYYYY...
# ..YYYYYYY..
# .YYYYYYYYY.
# .YYYYrYYYY.
# .YYYYYYYYY.
# ..YYYYYYY..
# ...YYYYY...
# ....YYY....
# ...........

#--------------------------------------------------------------------------------------------------------------------------#
# Step 1: Determine the map size.
#--------------------------------------------------------------------------------------------------------------------------#

def getGridSize(argsList):

        global Details

	gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()
	saveSettings()

	# Determine the width.

	Width = getSelectedMapValue("Min Width:")
	MaxWidth = getSelectedMapValue("Max Width:")

	if (MaxWidth > 0 and MaxWidth != Width):
                DiffWidth = abs(MaxWidth - Width) + 1
                Width = min(Width, MaxWidth) + dice.get(DiffWidth, "Width")

        # Determine the height.

        Height = getSelectedMapValue("Min Height:")

        if Height == 0:
                Height = int((Width*3.0)/5.0)
                MaxHeight = int((Width*5.0)/8.0)
                if MaxHeight != Height:
                        DiffHeight = abs(MaxHeight - Height) + 1
                        Height = min(Height, MaxHeight) + dice.get(DiffHeight, "Height")
        else:
                MaxHeight = getSelectedMapValue("Max Height:")
                if (MaxHeight > 0 and MaxHeight != Height):
                        DiffHeight = abs(MaxHeight - Height) + 1
                        Height = min(Height, MaxHeight) + dice.get(DiffHeight, "Height")

        # If requested, make note of the options that have been used.

        Details = getSelectedMapValue("Print Map Details:")

	if Details:

                f = open('C:\Program Files\Firaxis Games\Map Options', 'w')

                s = 'Width: ' + repr(Width*4) + '\n'
                f.write(s)
                s = 'Height: ' + repr(Height*4) + '\n'
                f.write(s)

                f.close()

        # Return the width and height.
	
	return (Width, Height)

#--------------------------------------------------------------------------------------------------------------------------#
# Step 2: Distinguish between the different base terrain types - ocean, flatland, hill or peak.
#--------------------------------------------------------------------------------------------------------------------------#

def generatePlotTypes():

	global Details
        global width
        global height
        global Area
	global fillType
	global NewplotTypes
	global UpdatedOwnerLandMasses
	global LandMassTileCount

	cymap = CyMap()
	gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()
	width = cymap.getGridWidth()
	height = cymap.getGridHeight()

	landmass = 0
	continent = 0
	problem = 0
	check = False
	Area = width*height
	CanBeLandLists = []
	OwnerLandMasses = [-1]*(Area)
	UpdatedOwnerLandMasses = [-1]*(Area)
	OwnerContinents = [-1]*(Area)
	plotTypes = [fillType]*(Area)
	NewplotTypes = [fillType]*(Area)
	PlotAltitude = [0]*(Area)
	NewPlotAltitude = [0]*(Area)
	RestrictedHeight = int(height/8)
	RestrictedAreaOne = RestrictedHeight*width
	RestrictedAreaTwo = Area - 2*RestrictedAreaOne
	LandPlaced = 0
	RestoreLake = 0
	HillsPlaced = 0
	PeaksPlaced = 0

	# Find the minimum seperation between land masses.

	SeaWidth = getSelectedMapValue("Min Oceanic Seperation:")
	MaxSeaWidth = getSelectedMapValue("Max Oceanic Seperation:")

	if (MaxSeaWidth >= 0 and MaxSeaWidth != SeaWidth):
                DiffSeaWidth = abs(MaxSeaWidth - SeaWidth) + 1
                SeaWidth = min(SeaWidth, MaxSeaWidth) + dice.get(DiffSeaWidth, "SeaWidth")

        SeaCheckSet = range(-SeaWidth - 3, SeaWidth + 2)

        # Find the amount of sea water covering the map.

        PercentLand = getSelectedMapValue("Min Ocean:")
	MaxLand = getSelectedMapValue("Max Ocean:")

	if (MaxLand > 0 and MaxLand != PercentLand):
                DiffLand = abs(MaxLand - PercentLand) + 1
                PercentLand = min(PercentLand, MaxLand) + dice.get(DiffLand, "Land")

        LandFractionOne = (PercentLand*Area)/100.0
        
        # What type of coast lines are to be generated.

        CoastLines = getSelectedMapValue("Coastlines:")

        if CoastLines < 0:
                CoastLines = 2 + dice.get(3, "Standard, rougher, or rugged coastlines.")

	# Determine the percentage of land tiles which are hills.

	PercentHills = getSelectedMapValue("Hills:")
	VariantHills = dice.get(100, "Randomly generate the exact percentage of hills.")
	
	if PercentHills == -1:
                PercentHills = 12.5 + (VariantHills/50.0)
	elif PercentHills == -2:
                PercentHills = 14.5 + (VariantHills/50.0)
        elif PercentHills == -3:
                PercentHills = 16.5 + (VariantHills/33.0)
        elif PercentHills == -4:
                PercentHills = 19.5 + (VariantHills/33.0)
        elif PercentHills == -5:
                PercentHills = 22.5 + (VariantHills/33.0)
        elif PercentHills == -6:
                PercentHills = 25.5 + (VariantHills/33.0)
        elif PercentHills == -7:
                PercentHills = 28.5 + (VariantHills/33.0)
        elif PercentHills == -8:
                PercentHills = 31.5 + (VariantHills/33.0)
        elif PercentHills == -9:
                PercentHills = 12.5 + dice.get(33, "Between 12.5% and 34.5% of land is hills.")
        elif PercentHills == 0:
                PercentHills = 40 + dice.get(11, "Between 40% and 50% of land is hills.")

        # Determine the percentage of land tiles which are peaks.

	PercentPeaks = getSelectedMapValue("Peaks:")
	VariantPeaks = dice.get(100, "Randomly generate the exact percentage of peaks.")

	if PercentPeaks == -1:
                PercentPeaks = VariantPeaks/100.0
	elif PercentPeaks == -2:
                PercentPeaks = 1.0 + (VariantPeaks/100.0)
        elif PercentPeaks == -3:
                PercentPeaks = 2.0 + (VariantPeaks/50.0)
        elif PercentPeaks == -4:
                PercentPeaks = 4.0 + (VariantPeaks/50.0)
        elif PercentPeaks == -5:
                PercentPeaks = 6.0 + (VariantPeaks/50.0)
        elif PercentPeaks == -6:
                PercentPeaks = 8.0 + (VariantPeaks/50.0)
        elif PercentPeaks == -7:
                PercentPeaks = 10.0 + (VariantPeaks/50.0)
        elif PercentPeaks == -8:
                PercentPeaks = 12.0 + (VariantPeaks/50.0)
        elif PercentPeaks == -9:
                PercentPeaks = dice.get(15, "Between 0% and 14% of land is peaks.")
        elif PercentPeaks == 0:
                PercentPeaks = 20 + dice.get(11, "Between 20% and 30% of land is peaks.")

        # Are mountain ranges generated?.

        MountainRanges = getSelectedMapValue("Mountain Ranges:")

        if MountainRanges < 0:
                MountainRanges = 0 + dice.get(3, "No ranges, ranges, or all ranges.")

        # Are altitude changes slow or rapid?

        AltitudeChanges = getSelectedMapValue("Altitude Changes:")

        if AltitudeChanges < 0:
                AltitudeChanges = 0 + dice.get(2, "Altitudes; clumped or not clumped.")
        
        # Find the number of land masses.

        LandMassCount = getSelectedMapValue("Min Land Masses:")
	MaxLandMassCount = getSelectedMapValue("Max Land Masses:")

	if (MaxLandMassCount > 0 and MaxLandMassCount != LandMassCount):
                DiffLandMassCount = abs(MaxLandMassCount - LandMassCount) + 1
                LandMassCount = min(LandMassCount, MaxLandMassCount) + dice.get(DiffLandMassCount, "Land Masses")

        LandMassTileCount = [0]*LandMassCount

        # Find the number of continents.

	ContinentCount = getSelectedMapValue("Min Continents:")

	if SeaWidth <= 1 or ContinentCount == 0:
                ContinentCount = LandMassCount
        else:
                MaxContinentCount = getSelectedMapValue("Max Continents:")

                if (MaxContinentCount > 0 and MaxContinentCount != ContinentCount):
                        DiffContinentCount = abs(MaxContinentCount - ContinentCount) + 1
                        ContinentCount = min(ContinentCount, MaxContinentCount) + dice.get(DiffContinentCount, "Continents")

	if ContinentCount < LandMassCount:
                ContinentCount = LandMassCount

        ContinentLandMass = [-1]*ContinentCount
        LandFractionTwo = max(LandFractionOne/(2*LandMassCount), SeaWidth**2)
        LandFractionThree = LandFractionOne/(2*ContinentCount)
        LandFractionFour = LandFractionOne/(3*ContinentCount)

        # Find the continent distribution.

        ContinentDistribution = getSelectedMapValue("Continent Distribution:")

        if ContinentDistribution < 0:
                ContinentDistribution = dice.get(2, "True or false.")

        # Find the land mass shape.

        LandMassShape = getSelectedMapValue("Continent Placement:")

        if LandMassShape < 0:
                LandMassShape = dice.get(3, "Rounded, snaky, or normal land masses.")
        
        # Begin land mass generation.

        while (continent < ContinentCount):
                continent += 1
		CanBeLandLists.append([])

	ExpectedLandPlotCount = int(LandFractionOne)
	LandToPlace = int((102*ExpectedLandPlotCount)/100)
	actualLandPlotCount = ContinentCount

	# If requested, make note of the options that have been used.

	if Details:

                f = open('C:\Program Files\Firaxis Games\Map Options', 'a')
                
                s = 'Land masses: ' + repr(LandMassCount) + '\n'
                f.write(s)

                s = 'Continents: ' + repr(ContinentCount) + '\n'
                f.write(s)

                f.close()

        # Generate a starting plot for each continent.

	# If only one continent is to be placed.

	if ContinentCount == 1:
                I = (int(height/2))*width + int(width/2)
                OwnerLandMasses[I] = landmass
                OwnerContinents[I] = continent
                LandMassTileCount[landmass] += 1
                plotTypes[I] = PlotTypes.PLOT_LAND
                PlotAltitude[I] = 1 + dice.get(100, "Altitude.")
                for i in nearDirs:
                        Inew = GetNewTileIndex(I, i)
                        CanBeLandLists[0].append(Inew)

	# If multiple continents are to be placed.

	elif ContinentCount > 1:

                ContinentTiles = []
                TotalDistances = []

                # Place the first continent.

                I = RestrictedAreaOne + dice.get(RestrictedAreaTwo, "The first land tile.")
                plotTypes[I] = PlotTypes.PLOT_LAND
                OwnerLandMasses[I] = 0
                OwnerContinents[I] = 0
                ContinentLandMass[0] = 0
                LandMassTileCount[landmass] += 1
                PlotAltitude[I] = 1 + dice.get(100, "Altitude.")
                ContinentTiles.append(I)
                for i in nearDirs:
                        Inew = GetNewTileIndex(I, i)
                        CanBeLandLists[0].append(Inew)

                # Determine the starting plot of each other continent.

                for continent in range(1, ContinentCount):

                        continent -= problem
                        PossibleStartTiles = []

                        # Which land mass is the continent part of?

                        if LandMassCount > continent:
                                landmass = continent
                        elif LandMassCount <= continent:
                                if ContinentDistribution:
                                        landmass = continent
                                        while landmass >= LandMassCount:
                                                landmass -= LandMassCount
                                else:
                                        landmass = dice.get(LandMassCount, "Pick a random land mass.")

                        ContinentLandMass[continent] = landmass

                        # The first tile of the last land mass is to be placed as far from other land as possible.

                        if continent == LandMassCount - 1:
                                FurthestDistance = -1
                                ShortestDistance = Area
                                for I in range(RestrictedAreaOne, RestrictedAreaTwo):
                                        conflict = False
                                        Distance = 0
                                        TotalDistances.append(0)
                                        for Inew in ContinentTiles:
                                                if I == Inew:
                                                        conflict = True
                                                        TotalDistances[I - RestrictedAreaOne] = 0
                                                        break
                                                X, Y = GetDistance(I, Inew)
                                                Distance = X**2 + Y**2
                                                if Distance >= LandFractionTwo:
                                                        TotalDistances[I - RestrictedAreaOne] += Distance
                                                elif Distance < LandFractionTwo:
                                                        conflict = True
                                                        TotalDistances[I - RestrictedAreaOne] = 0
                                                        break
                                        
                                        if not conflict:
                                                if TotalDistances[I - RestrictedAreaOne] > FurthestDistance:
                                                        FurthestDistance = TotalDistances[I - RestrictedAreaOne]
                                                if TotalDistances[I - RestrictedAreaOne] < ShortestDistance:
                                                        ShortestDistance = TotalDistances[I - RestrictedAreaOne]

                                # At what approximate distance from other land masses will the last land mass be placed?
                                
                                if FurthestDistance > 0:
                                        FurthestDistance = (FurthestDistance + ShortestDistance)/2

                                # Assign the starting point of the last land mass.

                                if FurthestDistance > 0:

                                        # Find a plot at the appropriate distance from other land masses.

                                        for I in range(len(TotalDistances)):
                                                TotalDistances[I] -= FurthestDistance
                                                if TotalDistances[I] < 0:
                                                        TotalDistances[I] -= 2*TotalDistances[I]
                                        
                                        FurthestDistance = min(TotalDistances)
                                        PossibleTiles = []

                                        for I in range(len(TotalDistances)):
                                                if TotalDistances[I] == FurthestDistance:
                                                        PossibleTiles.append(I)

                                        I = PossibleTiles[dice.get(len(PossibleTiles), "")] - 1 + RestrictedAreaOne

                                        # Place the land mass starting tile.
                                        
                                        plotTypes[I] = PlotTypes.PLOT_LAND
                                        OwnerLandMasses[I] = landmass
                                        OwnerContinents[I] = continent
                                        LandMassTileCount[landmass] += 1
                                        PlotAltitude[I] = 1 + dice.get(100, "Altitude.")
                                        ContinentTiles.append(I)
                                        for i in nearDirs:
                                                Inew = GetNewTileIndex(I, i)
                                                CanBeLandLists[continent].append(Inew)

                                # If the last land mass cannot be placed.

                                elif FurthestDistance <= 0:
                                        problem += 1
                                        LandMassCount -= 1
                                        ContinentCount -= 1
                                        del CanBeLandLists[continent]
                                        del ContinentLandMass[continent]
                                        del LandMassTileCount[landmass]
                                        actualLandPlotCount -= 1

                                continue

                        # If more rounded land masses or more snaky land masses have been selected.

                        if LandMassShape > 0:
                                landmasssize = 0
                                for continents in range(continent):
                                        if ContinentLandMass[continents] == landmass:
                                                landmasssize += 1

                        # If more rounded land masses has been selected.

                        if LandMassShape == 1 and landmasssize >= 2:
                                NearestDistance = Area
                                for I in range(RestrictedAreaOne, RestrictedAreaTwo):
                                        conflict = False
                                        MinDistance = -1
                                        TotalDistance = 0
                                        for Inew in ContinentTiles:
                                                if I == Inew:
                                                        conflict = True
                                                        break
                                                X, Y = GetDistance(I, Inew)
                                                Distance = X**2 + Y**2
                                                if OwnerLandMasses[Inew] == landmass:
                                                        TotalDistance += Distance
                                                        if MinDistance < 0:
                                                                MinDistance = Distance
                                                        else:
                                                                MinDistance = min(MinDistance, Distance)
                                                elif OwnerLandMasses[Inew] != landmass:
                                                        if Distance < LandFractionTwo:
                                                                conflict = True
                                                                break

                                        if conflict:
                                                continue

                                        if MinDistance > LandFractionThree:
                                                continue

                                        if MinDistance < LandFractionFour and MinDistance > 0:
                                                continue

                                        if TotalDistance < NearestDistance:
                                                NearestDistance = TotalDistance
                                                Ikept = I

                                # Assign the starting tile of the continent.

                                if FurthestDistance > 0:
                                        plotTypes[Ikept] = PlotTypes.PLOT_LAND
                                        OwnerLandMasses[Ikept] = landmass
                                        OwnerContinents[Ikept] = continent
                                        LandMassTileCount[landmass] += 1
                                        PlotAltitude[Ikept] = 1 + dice.get(100, "Altitude.")
                                        ContinentTiles.append(Ikept)
                                        for i in nearDirs:
                                                Inew = GetNewTileIndex(Ikept, i)
                                                CanBeLandLists[continent].append(Inew)

                                # If the continent cannot be placed.

                                elif FurthestDistance <= 0:
                                        problem += 1
                                        ContinentCount -= 1
                                        del CanBeLandLists[continent]
                                        del ContinentLandMass[continent]
                                        actualLandPlotCount -= 1

                                continue

                        # If more snaky land masses has been selected.

                        if LandMassShape == 2 and landmasssize >= 2:
                                FurthestDistance = -1
                                for I in range(RestrictedAreaOne, RestrictedAreaTwo):
                                        conflict = False
                                        MinDistance = -1
                                        TotalDistance = 0
                                        for Inew in ContinentTiles:
                                                if I == Inew:
                                                        conflict = True
                                                        break
                                                X, Y = GetDistance(I, Inew)
                                                Distance = X**2 + Y**2
                                                if OwnerLandMasses[Inew] == landmass:
                                                        TotalDistance += Distance
                                                        if MinDistance < 0:
                                                                MinDistance = Distance
                                                        else:
                                                                MinDistance = min(MinDistance, Distance)
                                                elif OwnerLandMasses[Inew] != landmass:
                                                        if Distance < LandFractionTwo:
                                                                conflict = True
                                                                break

                                        if conflict:
                                                continue

                                        if MinDistance > LandFractionThree:
                                                continue

                                        if MinDistance < LandFractionFour and MinDistance > 0:
                                                continue

                                        if TotalDistance > FurthestDistance:
                                                FurthestDistance = TotalDistance
                                                Ikept = I

                                # Assign the starting tile of the continent.

                                if FurthestDistance > 0:
                                        plotTypes[Ikept] = PlotTypes.PLOT_LAND
                                        OwnerLandMasses[Ikept] = landmass
                                        OwnerContinents[Ikept] = continent
                                        LandMassTileCount[landmass] += 1
                                        PlotAltitude[Ikept] = 1 + dice.get(100, "Altitude.")
                                        ContinentTiles.append(Ikept)
                                        for i in nearDirs:
                                                Inew = GetNewTileIndex(Ikept, i)
                                                CanBeLandLists[continent].append(Inew)

                                # If the last continent cannot be placed.

                                elif FurthestDistance <= 0:
                                        problem += 1
                                        ContinentCount -= 1
                                        del CanBeLandLists[continent]
                                        del ContinentLandMass[continent]
                                        actualLandPlotCount -= 1

                                continue

                        # Determine all possible starting tiles for continents that don't meet the above conditions.

                        for I in range(RestrictedAreaOne, RestrictedAreaTwo):
                                conflict = False
                                MinDistance = -1
                                for Inew in ContinentTiles:
                                        if I == Inew:
                                                conflict = True
                                                break
                                        X, Y = GetDistance(I, Inew)
                                        Distance = X**2 + Y**2
                                        if OwnerLandMasses[Inew] == landmass:
                                                if MinDistance < 0:
                                                        MinDistance = Distance
                                                else:
                                                        MinDistance = min(MinDistance, Distance)
                                        elif OwnerLandMasses[Inew] != landmass:
                                                if Distance < LandFractionTwo:
                                                        conflict = True
                                                        break

                                if conflict:
                                        continue

                                if MinDistance > LandFractionThree:
                                        continue

                                if MinDistance < LandFractionFour and MinDistance > 0:
                                        continue

                                PossibleStartTiles.append(I)

                        # Select the continent's starting tile.

                        if len(PossibleStartTiles) > 0:
                                I = PossibleStartTiles[dice.get(len(PossibleStartTiles), "Pick a possible starting tile.")]
                                plotTypes[I] = PlotTypes.PLOT_LAND
                                OwnerLandMasses[I] = landmass
                                OwnerContinents[I] = continent
                                LandMassTileCount[landmass] += 1
                                PlotAltitude[I] = 1 + dice.get(100, "Altitude.")
                                ContinentTiles.append(I)
                                for i in nearDirs:
                                        Inew = GetNewTileIndex(I, i)
                                        CanBeLandLists[continent].append(Inew)

                        # If no starting position is available.

                        else:

                                problem += 1
                                if LandMassCount > continent:
					LandMassCount -= 1
					del LandMassTileCount[landmass]
                                ContinentCount -= 1
                                del CanBeLandLists[continent]
                                del ContinentLandMass[continent]
                                actualLandPlotCount -= 1

        # If continents are to form mountain ranges, determine which continents do so.

        Ranges = []

        if MountainRanges:
                if MountainRanges == 2:
                        Ranges = range(ContinentCount)
                else:
                        RangeCount = 3 + dice.get(3, "How many continents produce ranges?")
                        if RangeCount >= ContinentCount:
                                Ranges = range(ContinentCount)
                        else:
                                while len(Ranges) < RangeCount:
                                        for continent in range(ContinentCount):
                                                if continent not in Ranges:
                                                        if dice.get(ContinentCount, "") >= RangeCount:
                                                                Ranges.append(continent)
                                                                if len(Ranges) == RangeCount:
                                                                        break

        # If requested, make note of the options that have been used.

        if Details:

                f = open('C:\Program Files\Firaxis Games\Map Options', 'a')

                s = 'Actual land masses: ' + repr(LandMassCount) + '\n'
                f.write(s)

                s = 'Actual continents: ' + repr(ContinentCount) + '\n'
                f.write(s)

                if ContinentDistribution:
                        f.write('Continent distribution: Evenly distributed\n')
                else:
                        f.write('Continent distribution: Randomly distributed\n')

                if LandMassShape == 0:
                        f.write('Continent placement: Normal\n')
                elif LandMassShape == 1:
                        f.write('Continent placement: More rounded\n')
                else:
                        f.write('Continent placement: More snaky\n')

                s = 'Minimum seperation between land masses: ' + repr(SeaWidth) + '\n'
                f.write(s)

                s = 'Ocean percentage: ' + repr(100 - PercentLand) + '\n'
                f.write(s)

                f.close()

        # Generate the land plots for each continent.

        tries = 0

        while (LandPlaced < CoastLines*LandToPlace/2.0 and tries < 4*CoastLines*LandToPlace):
                for continent in range(ContinentCount):

                        if ContinentCount > 1:
                                if check:
                                        check = False
                                        continent -= 1
                                        if continent < 0:
                                                continent = ContinentCount - 1
                                        landmass = ContinentLandMass[continent]
                                else:
                                        landmass = ContinentLandMass[continent]
                                        if LandMassTileCount[landmass] > 3.0*sum(LandMassTileCount)/(2 + 2*LandMassCount):
                                                continue

                        tries += 1
                        dropProb = 0
                        anyMatch = False

                        # Select a tile to be added to the continent.

                        if CanBeLandLists[continent] == []:
                                continue
                        tile = dice.get(len(CanBeLandLists[continent]), "Select a tile to be added to the continent.")
                        I = CanBeLandLists[continent][tile]

                        # Determine if the tile is too close to another land mass.

                        if (LandMassCount > 1 and not anyMatch):
                                Limit = max(SeaWidth - 2, 2)
                                for i in SeaCheckSet:
                                        for j in SeaCheckSet:
                                                if abs(i) + abs(j) < Limit:
                                                        continue
                                                Inew = GetNewTileIndex(I, (i, j))
                                                if Inew != -1:
                                                        if plotTypes[Inew] == PlotTypes.PLOT_OCEAN:
                                                                continue
                                                        if OwnerLandMasses[Inew] == landmass:
                                                                continue
                                                        X, Y = GetDistance(I, Inew)
                                                        if (X + Y) <= SeaWidth or (X <= SeaWidth and Y <= SeaWidth):
                                                                anyMatch = True
                                                                dropProb = 10001
                                                                break
                                                        sep = int(((X*X) + (Y*Y))**0.5)
                                                        beyondSeaWidthTest = 2000 - (100*(sep - SeaWidth)*(sep - SeaWidth))
                                                        edgeVal = dice.get(10000, "Random Continent sea edge bending")
                                                        if edgeVal < beyondSeaWidthTest:
                                                                dropProb = int(beyondSeaWidthTest/2)
                                                                anyMatch = True
                                                                break
                                                if anyMatch:
                                                        break

                        # Determine if the tile is too close to the poles.

                        if not anyMatch:
                                if I > Area/2:
                                        IndicesFromEnd = Area - I - 1
                                elif I <= Area/2:
                                        IndicesFromEnd = I
                                (X, Y) = GetDistance(IndicesFromEnd, 0)
                                Y = Y - 2
                                if Y < RestrictedHeight:
                                        Land = dice.get(10000, "Will the tile be land?")
                                        dropProb = (10001*(RestrictedHeight - Y))/RestrictedHeight
                                        if Land < dropProb:
                                                anyMatch = True

                        # If the land tile will not be placed.

                        if anyMatch:
                                drop = dice.get(10000, "Random canbeland drop")
                                if drop < dropProb:
                                        del CanBeLandLists[continent][tile]
                                if not check:
                                        check = True
                                        tries -= 1
                                elif check:
                                        check = False
                                continue

                        # If the land tile will be placed.
                        
                        del CanBeLandLists[continent][tile]

                        if plotTypes[I] == PlotTypes.PLOT_LAND:
                                PlotAltitude[I] += dice.get(100, "Additional altitude.")
                        elif plotTypes[I] == PlotTypes.PLOT_OCEAN:
                                plotTypes[I] = PlotTypes.PLOT_LAND
                                OwnerLandMasses[I] = landmass
                                OwnerContinents[I] = continent
                                LandMassTileCount[landmass] += 1
                                PlotAltitude[I] = 1 + dice.get(100, "Altitude.")
                                LandPlaced += 1

                                # Determine which tiles the continent can now incorporate.

                                for i in nearDirs:
                                        Inew = GetNewTileIndex(I, i)
                                        if Inew != -1:
                                                if not (Inew in CanBeLandLists[continent]):
                                                        if OwnerContinents[Inew] != continent:
                                                                if plotTypes[Inew] == PlotTypes.PLOT_LAND and PlotAltitude[Inew] <= 100 and continent in Ranges:
                                                                        PlotAltitude[Inew] += dice.get(100, "Additional Altitude.")
                                                                elif plotTypes[Inew] == PlotTypes.PLOT_OCEAN:
                                                                        roundTest = dice.get(1000, "Continent growth.")
                                                                        if roundTest < 301:
                                                                                CanBeLandLists[continent].append(Inew)
	
        actualLandPlotCount += LandPlaced

        # Fill in any freshwater lakes.

	for I in range(Area):

		if plotTypes[I] == PlotTypes.PLOT_OCEAN:
			continue

		# Is the land tile adjacent to an ocean tile? If so, determine if it is fresh water or salt water.

		if I > Area/2:
			Inew = GetNewTileIndex(I, cardinalDirs[1])
			if Inew == -1:
                                continue
		elif I <= Area/2:
			Inew = GetNewTileIndex(I, cardinalDirs[2])
			if Inew == -1:
                                continue

		if plotTypes[Inew] == PlotTypes.PLOT_OCEAN:
			oceancount = 1
			tilesused = 0
			LakesTiles = []
			LakesTiles.append(Inew)

			while oceancount < 10:

				tilesused += 1

				for i in cardinalDirs:
					Inext = GetNewTileIndex(Inew, i)
					if Inext != -1:
                                                if plotTypes[Inext] == PlotTypes.PLOT_OCEAN and not (Inext in LakesTiles):
                                                        LakesTiles.append(Inext)
                                                        oceancount += 1

				if oceancount >= 10:
					break

				if tilesused < len(LakesTiles):
					Inew = LakesTiles[tilesused]
				elif tilesused == len(LakesTiles):
					RestoreLake += tilesused
					landmass = OwnerLandMasses[I]
					for j in range(tilesused):
						Inext = LakesTiles[j]
						plotTypes[Inext] = PlotTypes.PLOT_LAND
						LandMassTileCount[landmass] += 1
						PlotAltitude[Inext] = 1
						OwnerLandMasses[Inext] = landmass
					break

	actualLandPlotCount += RestoreLake
        
	# Adjust the amount of ocean tiles to correct the percentage.
	
	landRemovalTries = 0

	while actualLandPlotCount > ExpectedLandPlotCount and landRemovalTries < ExpectedLandPlotCount * 2048:
                
		landRemovalTries += 1

		# Select a land tile for removal.
		
		I = dice.get(Area, "Tile proposed for removal.")
		
		if plotTypes[I] != PlotTypes.PLOT_LAND:
			continue

		if LandMassTileCount[OwnerLandMasses[I]] < 1.0*sum(LandMassTileCount)/(LandMassCount + 1):
                        continue

		# Determine if the tile is adjacent to a coastal tile.
		
		for i in cardinalDirs:
			Inew = GetNewTileIndex(I, i)
			if Inew != -1 and plotTypes[Inew] == fillType:

                                # Set the tile to water.
                                
                                plotTypes[I] = fillType
                                PlotAltitude[I] = 0
                                actualLandPlotCount -= 1
                                landmass = OwnerLandMasses[I]
                                LandMassTileCount[landmass] -= 1
                                OwnerLandMasses[I] = -1

                                # The odds of archipelagos forming increases as the ruggedness of the coast increases.

                                j = dice.get(100, "A random number.")

                                while j < (CoastLines - 1)*25:
                                        
                                        if actualLandPlotCount <= ExpectedLandPlotCount:
                                                j = 100

                                        else:
                                                AdjacentLandMassTiles = []
                                                for i in cardinalDirs:
                                                        Inew = GetNewTileIndex(I, i)
                                                        if Inew != -1 and plotTypes[Inew] == PlotTypes.PLOT_LAND:
                                                                AdjacentLandMassTiles.append(Inew)

                                                # If there are adjacent land tiles, set one to be water.

                                                if len(AdjacentLandMassTiles) > 0:
                                                        I = AdjacentLandMassTiles[dice.get(len(AdjacentLandMassTiles), "")]
                                                        plotTypes[I] = fillType
                                                        PlotAltitude[I] = 0
                                                        actualLandPlotCount -= 1
                                                        LandMassTileCount[landmass] -= 1
                                                        OwnerLandMasses[I] = -1
                                                        j = dice.get(100, "A random number.")

                                                else:
                                                        j = 100

                                break

        # If requested, make note of the options that have been used.

        if Details:

                f = open('C:\Program Files\Firaxis Games\Map Options', 'a')

                s = 'Actual ocean percentage: ' + repr((Area - actualLandPlotCount)*100/Area) + '\n'
                f.write(s)

                if CoastLines == 2:
                        f.write('Coast lines: Standard\n')
                elif CoastLines == 3:
                        f.write('Coast lines: Rougher\n')
                else:
                        f.write('Coast lines: Rugged\n')

                s = 'Hills percentage: ' + repr(PercentHills) + '\n'
                f.write(s)

                s = 'Peaks percentage: ' + repr(PercentPeaks) + '\n'
                f.write(s)
                
                if MountainRanges == 0:
                        f.write('Mountain ranges: None\n')
                elif MountainRanges == 1:
                        s = 'Mountain ranges: Ranges at the borders of ' + repr(RangeCount) + ' continents.\n'
                        f.write(s)
                else:
                        f.write('Mountain ranges: Ranges at the borders of all continents\n')
                if AltitudeChanges == 0:
                        f.write('Placement of hills and peaks unaffected by altitude of neighbouring tiles.\n')
                else:
                        f.write('Placement of hills and peaks affected by altitude of neighbouring tiles.\n')

                f.close()

	# Alter the rate at which altitude changes.

	if AltitudeChanges:
                for I in range(Area):

                        if plotTypes[I] == PlotTypes.PLOT_OCEAN:
                                continue

                        LandTile = 0
                        NearAltitude = 0

                        for i in nearDirs:
                                Inext = GetNewTileIndex(I, i)
                                if Inext != -1:
                                        if plotTypes[Inext] == PlotTypes.PLOT_LAND:
                                                LandTile += 1
                                                NearAltitude += PlotAltitude[Inext]

                        if LandTile > 0:
                                NewPlotAltitude[I] = int((PlotAltitude[I] + NearAltitude/LandTile)/2)
                        else:
                                NewPlotAltitude[I] = PlotAltitude[I]

        else:
                for I in range(Area):
                        NewPlotAltitude[I] = PlotAltitude[I]
                                        
        
	# Place peaks.

	RequiredPeakCount = int((PercentPeaks*actualLandPlotCount)/100)

	while PeaksPlaced < RequiredPeakCount:
		
		MaxAltitude = max(NewPlotAltitude)
		NewPeaks = []

                for I in range(Area):
                        if NewPlotAltitude[I] == MaxAltitude:
                                plotTypes[I] = PlotTypes.PLOT_PEAK
                                NewPeaks.append([I])
                                NewPlotAltitude[I] = 0
                                PeaksPlaced += 1

        # Remove excess peaks.

        if PeaksPlaced > RequiredPeakCount:
		
                ExcessPeaks = PeaksPlaced - RequiredPeakCount
                
                for N in range(ExcessPeaks):
                        I = dice.get(len(NewPeaks), "Pick a smaller peak.")
                        for J in range(Area):
                                if J in NewPeaks[I]:
                                        plotTypes[J] = PlotTypes.PLOT_HILLS
                                        del NewPeaks[I]
                                        PeaksPlaced -= 1
                                        HillsPlaced += 1
                                        break

        # Place hills.

        RequiredHillCount = int((PercentHills*actualLandPlotCount)/100)

        while HillsPlaced < RequiredHillCount:

                MaxAltitude = max(NewPlotAltitude)
                NewHills = []

                for I in range(Area):
                        if NewPlotAltitude[I] == MaxAltitude:
                                plotTypes[I] = PlotTypes.PLOT_HILLS
                                NewHills.append([I])
                                NewPlotAltitude[I] = 0
                                HillsPlaced += 1

        # Remove excess hills.

        if HillsPlaced > RequiredHillCount:
		
                ExcessHills = HillsPlaced - RequiredHillCount
                
                for N in range(ExcessHills):
                        I = dice.get(len(NewHills), "Pick a smaller hill.")
                        for J in range(Area):
                                if J in NewHills[I]:
                                        plotTypes[J] = PlotTypes.PLOT_LAND
                                        del NewHills[I]
                                        HillsPlaced -= 1
                                        break

	# Place the widest ocean on the x-wrap.

	ConsecutiveOceanStrips = []
	CurrentConsecutiveOceanStrips = []

	for X in range(width):
                for Y in range(height):
                        I = Y*width + X
                        if plotTypes[I] != PlotTypes.PLOT_OCEAN:
                                CurrentConsecutiveOceanStrips = []
                                break
                        if Y == height - 1:
                                CurrentConsecutiveOceanStrips.append(X)
                if len(CurrentConsecutiveOceanStrips) > len(ConsecutiveOceanStrips):
                        ConsecutiveOceanStrips = []
                        ConsecutiveOceanStrips = CurrentConsecutiveOceanStrips

        if len(ConsecutiveOceanStrips) > 0:
                Shift = ConsecutiveOceanStrips[int(len(ConsecutiveOceanStrips)/2)]
                for X in range(width):
                        for Y in range(height):
                                I = Y*width + X
                                if X >= Shift:
                                        Inew = I - Shift
                                        UpdatedOwnerLandMasses[Inew] = OwnerLandMasses[I]
                                        NewplotTypes[Inew] = plotTypes[I]
                                else:
                                        Inew = I + width - Shift
                                        UpdatedOwnerLandMasses[Inew] = OwnerLandMasses[I]
                                        NewplotTypes[Inew] = plotTypes[I]

        else:
                UpdatedOwnerLandMasses = OwnerLandMasses
                NewplotTypes = plotTypes
	
	# Finish up.

	return NewplotTypes

#--------------------------------------------------------------------------------------------------------------------------#
# Step 3: Distinguish between the different fine terrain types.
#--------------------------------------------------------------------------------------------------------------------------#

def generateTerrainTypes():

        global Details

        cymap = CyMap()
	gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()

        Climate = getSelectedMapValue("Climate:")
        Desert = getSelectedMapValue("Desert:")
        Plains = getSelectedMapValue("Plains:")
        TerrainHomogeneity = getSelectedMapValue("Terrain Clumping:")

        # If 'Random' has been selected for any of the options.

        if Climate == -1:
                Climate = dice.get(3, "Cool, temperate, or warm.")

        if Desert == 1:
                Desert = (1 + dice.get(3, "Less, normal, or more desert."))*16

        if Plains == 1:
                Plains = (1 + dice.get(3, "Less, normal, or more plains."))*9

        if TerrainHomogeneity == 1:
                TerrainHomogeneity = 2 + dice.get(5, "very varied, varied, normal, clumped, or very clumped.")

        # If requested, make note of the options that have been used.

        if Details:

                f = open('C:\Program Files\Firaxis Games\Map Options', 'a')
                
                if Climate < 0:
                        f.write('Fantasy terrain generated.\n')
                else:
                        if Climate == 0:
                                f.write('Temperature: Warm\n')
                        elif Climate == 1:
                                f.write('Temperature: Temperate\n')
                        else:
                                f.write('Temperature: Cool\n')
                        if Desert == 16:
                                f.write('Desert: Less than normal\n')
                        elif Desert == 32:
                                f.write('Desert: Normal\n')
                        else:
                                f.write('Desert: More than normal\n')
                        if Plains == 9:
                                f.write('Plains: Less than normal\n')
                        elif Plains == 18:
                                f.write('Plains: Normal\n')
                        else:
                                f.write('Plains: More than normal\n')
                        if TerrainHomogeneity == 2:
                                f.write('Terrain distribution: Very clumped\n')
                        elif TerrainHomogeneity == 3:
                                f.write('Terrain distribution: Clumped\n')
                        elif TerrainHomogeneity == 4:
                                f.write('Terrain distribution: Normal\n')
                        elif TerrainHomogeneity == 5:
                                f.write('Terrain distribution: Varied\n')
                        else:
                                f.write('Terrain distribution: Very varied\n')

                f.close()

        # How much snow and tundra is to be generated.

        if Climate == 0:
                SnowLatitude = 0.8
                TundraLatitude = 0.7
        elif Climate == 1:
                SnowLatitude = 0.7
                TundraLatitude = 0.6
        elif Climate == 2:
                SnowLatitude = 0.6
                TundraLatitude = 0.5

        TerrainGen = TerrainGenerator(Desert, Plains, SnowLatitude, TundraLatitude, 0.1, 0.2, 0.5, -1, -1,
                                      TerrainHomogeneity)
        TerrainTypes = TerrainGen.generateTerrain()
        return TerrainTypes

#--------------------------------------------------------------------------------------------------------------------------#
# Step 3A: Add rivers.
#--------------------------------------------------------------------------------------------------------------------------#

def addRivers():

        CyPythonMgr().allowDefaultImpl()
	return
	
#--------------------------------------------------------------------------------------------------------------------------#
# Step 3B: Add lakes.
#--------------------------------------------------------------------------------------------------------------------------#

def addLakes():

        CyPythonMgr().allowDefaultImpl()
	return

#--------------------------------------------------------------------------------------------------------------------------#
#Step 4: Add features.
#--------------------------------------------------------------------------------------------------------------------------#

def addFeatures():

        global Details

        cymap = CyMap()
	gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()

        Jungle = getSelectedMapValue("Jungle:")
        Forest = getSelectedMapValue("Forest:")
        FloodPlains = getSelectedMapValue("Flood Plains:")

        # If 'Random' has been selected for any options.

        if Jungle == 1:
                Jungle = 60 + (dice.get(3, "Less, normal, or more jungle."))*20

        if Forest == 1:
                Forest = 45 + (dice.get(3, "Less, normal, or more forest."))*15

        FeatureGen = FeatureGenerator(Jungle, Forest, 5, 6, -1, -1)
        FeatureTypes = FeatureGen.addFeatures()

        # If requested, make note of the options that have been used.

        if Details:

                f = open('C:\Program Files\Firaxis Games\Map Options', 'a')
                
                if Jungle == 60:
                        f.write('Jungle: More than normal\n')
                elif Jungle == 80:
                        f.write('Jungle: Normal\n')
                else:
                        f.write('Jungle: Less than normal\n')
                if Forest == 45:
                        f.write('Forest: More than normal\n')
                elif Forest == 60:
                        f.write('Forest: Normal\n')
                else:
                        f.write('Forest: Less than normal\n')
                if FloodPlains < 0:
                        f.write('Flood plains: Normal\n')
                else:
                        f.write('Flood plains: Alternative\n')

                f.close()

        # If the alternative flood plains generator has been selected.

        if FloodPlains == 2:

                global Area
                global NewplotTypes

                cymap = CyMap()
                gc = CyGlobalContext()

                for I in range(Area):

                        if NewplotTypes[I] != PlotTypes.PLOT_LAND:
                                continue
                        
                        Plot = cymap.plotByIndex(I)
                        Terrain = Plot.getTerrainType()
                        Features = Plot.getFeatureType()

                        if Features != -1:
                                continue

                        if not Plot.isRiverSide():
                                continue

                        if Terrain == 0 or Terrain == 1 or Terrain == 3:
                                if dice.get(3, "Place a flood plain?") == FloodPlains:
                                        Plot.setFeatureType(3, -1)

        return 0

#--------------------------------------------------------------------------------------------------------------------------#
# Step 5: Add bonuses.
#--------------------------------------------------------------------------------------------------------------------------#

def addBonuses():

        global Details
        
        cymap = CyMap()
	gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()

        # What resource generation type is used; standard or LandMasses?

        Resources = getSelectedMapValue("Resources:")

        # If standard ignore all other options.

        if not Resources:
		CyPythonMgr().allowDefaultImpl()

		f = open('C:\Program Files\Firaxis Games\Map Options', 'a')
		f.write('Resources: Standard\n')
		f.close()
		
		return

	# If LandMasses take account of selected options.
        
        ResourceRegion = getSelectedMapValue("Resource Land Masses:")
        ResourceIsolations = getSelectedMapValue("How Many Land Masses:")
	ResourceTiles = getSelectedMapValue("Resource Tiles:")
	StrategicQuantity = getSelectedMapValue("Strategic Supply:")
	HealthQuantity = getSelectedMapValue("Health Supply:")
	LuxuryQuantity = getSelectedMapValue("Luxury Supply:")
	ResourceClumps = getSelectedMapValue("Let Resources Clump:")

	# Determine the number of actual generated land masses.

	LandMasses = GetLandMasses()

	# If 'Random' has been selected for any options.

	if ResourceRegion < 0:
                ResourceRegion = dice.get(4, "Standard, LandMasses, LandMasses - isolate horses, or no isolated resources.")

        if ResourceIsolations == 0:
                ResourceIsolations = 1 + dice.get((len(LandMasses) - 1), "between 1 and (LandMasses - 1).")

        if StrategicQuantity != -5 and HealthQuantity != -5 and LuxuryQuantity != -5:

                if StrategicQuantity == 0:
                        StrategicQuantity = 1 + dice.get(3, "Less, normal, or more strategic quantities.")
                        if StrategicQuantity == 1:
                                StrategicQuantity = -1
                        if StrategicQuantity == 2:
                                StrategicQuantity = -2
                        if StrategicQuantity == 3:
                                StrategicQuantity = -3

                if HealthQuantity == 0:
                        HealthQuantity = 1 + dice.get(3, "Less, normal, or more health quantity")
                        if HealthQuantity == 1:
                                HealthQuantity = -1
                        if HealthQuantity == 2:
                                HealthQuantity = -2
                        if HealthQuantity == 3:
                                HealthQuantity = -3

                if LuxuryQuantity == 0:
                        LuxuryQuantity = 1 + dice.get(3, "Less, normal, or more luxury quantity")
                        if LuxuryQuantity == 1:
                                LuxuryQuantity = -1
                        if LuxuryQuantity == 2:
                                LuxuryQuantity = -2
                        if LuxuryQuantity == 3:
                                LuxuryQuantity = -3

	# If requested, make note of the options that have been used.

	if Details:

                f = open('C:\Program Files\Firaxis Games\Map Options', 'a')

                f.write('Resources: Custom\n')

                if ResourceRegion == 0:
                        f.write('Resource allocation: Standard\n')
                elif ResourceRegion == 1:
                        f.write('Resource allocation: LandMasses\n')
                elif ResourceRegion == 2:
                        f.write('Resource allocation: LandMasses - isolate horses\n')
                elif ResourceRegion == 3:
                        f.write('Resource allocation: No isolated resources\n')
                else:
                        f.write('Resource allocation: Custom\n')

                if ResourceIsolations == -1:
                        f.write('Isolated resources are assigned to half of the land masses.\n')
                elif ResourceIsolations == -2:
                        f.write('Isolated resources are assigned to all but one of the land masses.\n')
                elif ResourceIsolations == -3:
                        f.write('Isolated resources are assigned to land masses using a customised method.\n')
                else:
                        s = 'Isolated resources are assigned to ' + repr(ResourceIsolations) + ' land masses.\n'
                        f.write(s)

                if ResourceTiles == 0:
                        f.write('Resource tiles: Standard\n')
                elif ResourceTiles == 1:
                        f.write('Resource tiles: Land Masses\n')
                elif ResourceTiles == 3:
                        f.write('Resource tiles: Crazy\n')
                else:
                        f.write('Resource tiles: Custom\n')

                if StrategicQuantity == -5 or HealthQuantity == -5 or LuxuryQuantity == -5:
                        f.write('Strategic quantities: Custom\n')
                        f.write('Health quantities: Custom\n')
                        f.write('Luxury quantities: Custom\n')
                else:
                        if StrategicQuantity == -1:
                                f.write('Strategic quantities: Less than normal\n')
                        elif StrategicQuantity == -2:
                                f.write('Strategic quantities: Normal\n')
                        elif StrategicQuantity == -3:
                                f.write('Strategic quantities: More than normal\n')
                        else:
                                f.write('Strategic quantities: 1 per civilization\n')
                        if HealthQuantity == -1:
                                f.write('Health quantities: Less than normal\n')
                        elif HealthQuantity == -2:
                                f.write('Health quantities: Normal\n')
                        elif HealthQuantity == -3:
                                f.write('Health quantities: More than normal\n')
                        else:
                                f.write('Health quantities: 1 per civilization\n')
                        if LuxuryQuantity == -1:
                                f.write('Luxury quantities: Less than normal\n')
                        elif LuxuryQuantity == -2:
                                f.write('Luxury quantities: Normal\n')
                        elif LuxuryQuantity == -3:
                                f.write('Luxury quantities: More than normal\n')
                        else:
                                f.write('Luxury quantities: 1 per civilization\n')

                f.close()

	global Area
	global NewplotTypes
	global PlayerStartCross

	Players = gc.getGame().countCivPlayersEverAlive()
	FeatureOasis = gc.getInfoTypeForString("FEATURE_OASIS")
	FeatureIce = gc.getInfoTypeForString("FEATURE_ICE")

	if ResourceTiles == 1:

                TerrainGrass = gc.getInfoTypeForString("TERRAIN_GRASS")
                TerrainPlain = gc.getInfoTypeForString("TERRAIN_PLAINS")
                TerrainDesert = gc.getInfoTypeForString("TERRAIN_DESERT")
                TerrainTundra = gc.getInfoTypeForString("TERRAIN_TUNDRA")
                TerrainSnow = gc.getInfoTypeForString("TERRAIN_SNOW")
                FeatureForest = gc.getInfoTypeForString("FEATURE_FOREST")
                FeatureJungle = gc.getInfoTypeForString("FEATURE_JUNGLE")
                FeatureFloodPlain = gc.getInfoTypeForString("FEATURE_FLOOD_PLAINS")

                # AllLand contains the resources that can go on all land tiles.
                # AllWater contains the resources that can go on all water tiles.
                # Each other list contains the resources that can go on the specific tile type listed.

                AllLand = [0, 1, 2, 4, 5, 6, 7, 8, 22, 23, 24, 27]
                AllWater = [6, 10, 13, 15, 31]
                Snow = [21]
                HillSnow = [21]
                Tundra = [14, 18, 21, 25]
                ForestTundra = [14, 21, 25]
                HillTundra = [14, 18, 21]
                HillForestTundra = [14, 21]
                FloodPlainTundra = []
                Grass = [3, 11, 12, 16, 17, 18, 19, 20, 25, 28, 29, 30]
                ForestGrass = [9, 16, 20, 25, 26, 28, 29, 30]
                JungleGrass = [9, 16, 20, 25, 26, 28, 29, 30]
                HillGrass = [16, 18, 20, 28, 30]
                HillForestGrass = [9, 16, 20, 26, 28, 29, 30]
                HillJungleGrass = [9, 16, 20, 26, 28, 29, 30]
                FloodPlainGrass = []
                Plain = [3, 11, 12, 14, 16, 17, 18, 19, 20, 21, 25, 28, 29, 30]
                ForestPlain = [9, 14, 16, 20, 21, 25, 26, 28, 29, 30]
                HillPlain = [14, 16, 18, 20, 21, 28, 30]
                HillForestPlain = [9, 14, 16, 20, 21, 26, 28, 29, 30]
                FloodPlainPlain = []
                Desert = []
                HillDesert = []
                FloodPlainDesert = []

	elif ResourceTiles == 3:

                TerrainGrass = gc.getInfoTypeForString("TERRAIN_GRASS")
                TerrainPlain = gc.getInfoTypeForString("TERRAIN_PLAINS")
                TerrainDesert = gc.getInfoTypeForString("TERRAIN_DESERT")
                TerrainTundra = gc.getInfoTypeForString("TERRAIN_TUNDRA")
                TerrainSnow = gc.getInfoTypeForString("TERRAIN_SNOW")
                TerrainCoast = gc.getInfoTypeForString("TERRAIN_COAST")
                TerrainOcean = gc.getInfoTypeForString("TERRAIN_OCEAN")
                FeatureForest = gc.getInfoTypeForString("FEATURE_FOREST")
                FeatureJungle = gc.getInfoTypeForString("FEATURE_JUNGLE")
                FeatureFloodPlain = gc.getInfoTypeForString("FEATURE_FLOOD_PLAINS")

                (AllLand, AllWater, Coast, Ocean, Snow, HillSnow, Tundra, ForestTundra, HillTundra, HillForestTundra,
                 FloodPlainTundra, Grass, ForestGrass, JungleGrass, HillGrass, HillForestGrass, HillJungleGrass,
                 FloodPlainGrass, Plain, ForestPlain, HillPlain, HillForestPlain, FloodPlainPlain, Desert, HillDesert,
                 FloodPlainDesert) = CustomiseResourceTiles()

        else:

                # AllLand contains the resources that can go on all land tiles.
                # AllWater contains the resources that can go on all water tiles.

                AllLand = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 14, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30]
                AllWater = [6, 10, 13, 15, 31]

	# Isolated lists the resources that are unique to their respective landmasses.

	if ResourceRegion == 0:
                Isolated = [9, 11, 12, 14, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30]
        elif ResourceRegion == 1:
                Isolated = [11, 12, 16, 17, 18, 19, 25, 26, 29, 30]
        elif ResourceRegion == 2:
                Isolated = [3, 11, 12, 16, 17, 18, 19, 25, 26, 29, 30]
        elif ResourceRegion == 4:
                Isolated = CustomiseResourceLandMasses()

        # Determine how many land masses each resource is placed on.

        if ResourceIsolations > 0 and ResourceIsolations < len(LandMasses):
                Isolations = [ResourceIsolations]*32
        elif ResourceIsolations == -1 and len(LandMasses) > 1:
                Isolations = [len(LandMasses)/2]*32
        elif ResourceIsolations == -2 and len(LandMasses) > 1:
                Isolations = [len(LandMasses) - 1]*32
        elif ResourceIsolations == -3:
                Isolations = CustomisingHowManyLandMasses(len(LandMasses))
                for Resource in range(len(Isolations)):
                        if Isolations[Resource] > len(LandMasses):
                                Isolations[Resource] = len(LandMasses)
                        elif Isolations[Resource] == -1:
                                Isolations[Resource] = len(LandMasses)/2
                        elif Isolations[Resource] == -2:
                                Isolations[Resource] = len(LandMasses) - 1
                        elif Isolations[Resource] < -2:
                                Isolations[Resource] = len(LandMasses)
                        if Isolations[Resource] == 0:
                                Isolations[Resource] = len(LandMasses)
        else:
                ResourceRegion = 3

        # Set up the list containing the land masses each resource is assigned to.

        if ResourceRegion != 3:
                ResourceLM = []
                for Resource in range(32):
                        if Resource in Isolated and Isolations[Resource] > 0 and Isolations[Resource] < len(LandMasses):
                                ResourceLM.append([-1]*(Isolations[Resource]))
                        else:
                                ResourceLM.append(-1)
        else:
                ResourceLM = [-1]*(32)

        # Determine which land masses each resource is assigned to.

        if ResourceRegion != 3:
                for Resource in Isolated:
                        for LandMass in range(len(ResourceLM[Resource])):
                                while ResourceLM[Resource][LandMass] == -1:
                                        I = dice.get(Area, "Pick a tile.")
                                        for i in range(len(LandMasses)):
                                                if I in LandMasses[i]:
                                                        if i not in ResourceLM[Resource]:
                                                                ResourceLM[Resource][LandMass] = i
                                                        break
	
	# Determine all possible locations for each bonus.
	
	EligibleTiles = []
	
	for Resource in range(32):
		EligibleTiles.append([])
	
	for I in range(Area):

                # No resources are placed on peaks.

                if NewplotTypes[I] == PlotTypes.PLOT_PEAK:
                        continue

                Plot = cymap.plotByIndex(I)
                Terrain = Plot.getTerrainType()
                Feature = Plot.getFeatureType()

                # No resources placed under ice or on an oasis.

                if Feature == FeatureIce or Feature == FeatureOasis:
                        continue

                # Is the tile a part of any land mass?

                for i in range(len(LandMasses)):
                        if I in LandMasses[i]:
                                LandMass = i
                                break

                # Determine appropriate tiles for aquatic resources.

                if Plot.isWater():

                        # If the tile is not too far from flat land or a hill it can contain aquatic resources.
                        
                        for i in PlayerStartCross:
                                Inew = GetNewTileIndex(I, i)
                                if Inew == -1:
                                        continue
                                if NewplotTypes[Inew] == PlotTypes.PLOT_LAND or NewplotTypes[Inew] == PlotTypes.PLOT_HILLS:
                                        for Resource in AllWater:
                                                if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                        if ResourceTiles != 0:
                                                                EligibleTiles[Resource].append(I)
                                                        elif ResourceTiles == 0:
                                                                if Plot.canHaveBonus(Resource, True):
                                                                        EligibleTiles[Resource].append(I)
                                        if ResourceTiles == 3:
                                                if Terrain == TerrainCoast:
                                                        for Resource in Coast:
                                                                if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                        EligibleTiles[Resource].append(I)
                                                elif Terrain == TerrainOcean:
                                                        for Resource in Ocean:
                                                                if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                        EligibleTiles[Resource].append(I)
                                        break

                        continue

                # Determine which tiles are suitable for land resources.

                for Resource in AllLand:
                        
                        if ResourceLM[Resource] != -1 and LandMass not in ResourceLM[Resource]:
                                continue

                        if ResourceTiles != 0:
                                EligibleTiles[Resource].append(I)
                        elif ResourceTiles == 0:
                                if Plot.canHaveBonus(Resource, True):
                                        EligibleTiles[Resource].append(I)

                if ResourceTiles == 1 or ResourceTiles == 3:

                        if Terrain == TerrainSnow:
                                if NewplotTypes[I] == PlotTypes.PLOT_HILLS:
                                        for Resource in HillSnow:
                                                if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                        EligibleTiles[Resource].append(I)
                                elif NewplotTypes[I] == PlotTypes.PLOT_LAND:
                                        for Resource in Snow:
                                                if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                        EligibleTiles[Resource].append(I)

                        elif Terrain == TerrainTundra:
                                if NewplotTypes[I] == PlotTypes.PLOT_HILLS:
                                        if Feature == FeatureForest:
                                                for Resource in HillForestTundra:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        else:
                                                for Resource in HillTundra:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                elif NewplotTypes[I] == PlotTypes.PLOT_LAND:
                                        if Feature == FeatureForest:
                                                for Resource in ForestTundra:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        elif Feature == FeatureFloodPlain:
                                                for Resource in FloodPlainTundra:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        else:
                                                for Resource in Tundra:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)

                        elif Terrain == TerrainGrass:
                                if NewplotTypes[I] == PlotTypes.PLOT_HILLS:
                                        if Feature == FeatureForest:
                                                for Resource in HillForestGrass:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        elif Feature == FeatureJungle:
                                                for Resource in HillJungleGrass:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        else:
                                                for Resource in HillGrass:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                elif NewplotTypes[I] == PlotTypes.PLOT_LAND:
                                        if Feature == FeatureForest:
                                                for Resource in ForestGrass:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        elif Feature == FeatureJungle:
                                                for Resource in JungleGrass:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        elif Feature == FeatureFloodPlain:
                                                for Resource in FloodPlainGrass:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        else:
                                                for Resource in Grass:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)

                        elif Terrain == TerrainPlain:
                                if NewplotTypes[I] == PlotTypes.PLOT_HILLS:
                                        if Feature == FeatureForest:
                                                for Resource in HillForestPlain:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        else:
                                                for Resource in HillPlain:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                elif NewplotTypes[I] == PlotTypes.PLOT_LAND:
                                        if Feature == FeatureForest:
                                                for Resource in ForestPlain:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        elif Feature == FeatureFloodPlain:
                                                for Resource in FloodPlainPlain:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        else:
                                                for Resource in Plain:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)

                        elif Terrain == TerrainDesert:
                                if NewplotTypes[I] == PlotTypes.PLOT_HILLS:
                                        for Resource in HillDesert:
                                                if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                        EligibleTiles[Resource].append(I)
                                if NewplotTypes[I] == PlotTypes.PLOT_LAND:
                                        if Feature == FeatureFloodPlain:
                                                for Resource in FloodPlainDesert:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)
                                        else:
                                                for Resource in Desert:
                                                        if ResourceLM[Resource] == -1 or LandMass in ResourceLM[Resource]:
                                                                EligibleTiles[Resource].append(I)

        # Prepare to place the resources.

        ResourcesOrder = []
        ResourceLocations = []

        if StrategicQuantity == -5 or HealthQuantity == -5 or LuxuryQuantity == -5:
                ResourceQuantities = CustomiseResourceQuantities()
        
	for Resource in range(gc.getNumBonusInfos()):
		ResourceInfoXML = gc.getBonusInfo(Resource)
		if ResourceInfoXML.getPlacementOrder() >= 0:
			ResourcesOrder.append((Resource,ResourceInfoXML.getPlacementOrder()))

	ResourcesOrder = sorted(ResourcesOrder, key = operator.itemgetter(1))
	
	for ResourceOrder in ResourcesOrder:

                added = 0
                tries = 0
                NearLocations = []
                Resource, Order = ResourceOrder
                ResourceInfoXML = gc.getBonusInfo(Resource)

                # If modded resources are to be placed, LandMasses tries to place them in the standard way.

                if Resource >= 35:
			addBonusType([Resource])

		# Determine how much of the resource is to be placed.

		if Resource < 9:
                        if StrategicQuantity == -5:
                                ResourceCount = ResourceQuantities[Resource]
                                if ResourceCount <= -5:
                                        ResourceCount = -2
                                elif ResourceCount == -4:
                                        ResourceCount == Players
                                elif ResourceCount == -1:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                        ResourceCount = ResourceCount/2
                                elif ResourceCount == -2:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                elif ResourceCount == -3:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                        ResourceCount = ResourceCount*2
                        elif StrategicQuantity == -4:
                                ResourceCount = Players
                        elif StrategicQuantity > -4:
                                ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                if StrategicQuantity == -1:
                                        ResourceCount = ResourceCount/2
                                elif StrategicQuantity == -3:
                                        ResourceCount = ResourceCount*2
                elif Resource < 20:
                        if HealthQuantity == -5:
                                ResourceCount = ResourceQuantities[Resource]
                                if ResourceCount <= -5:
                                        ResourceCount = -2
                                elif ResourceCount == -4:
                                        ResourceCount == Players
                                elif ResourceCount == -1:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                        ResourceCount = ResourceCount/2
                                elif ResourceCount == -2:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                elif ResourceCount == -3:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                        ResourceCount = ResourceCount*2
                        elif HealthQuantity == -4:
                                ResourceCount = Players
                        elif HealthQuantity > -4:
                                ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                if HealthQuantity == -1:
                                        ResourceCount = ResourceCount/2
                                elif HealthQuantity == -3:
                                        ResourceCount = ResourceCount*2
                elif Resource < 32:
                        if LuxuryQuantity == -5:
                                ResourceCount = ResourceQuantities[Resource]
                                if ResourceCount <= -5:
                                        ResourceCount = -2
                                elif ResourceCount == -4:
                                        ResourceCount == Players
                                elif ResourceCount == -1:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                        ResourceCount = ResourceCount/2
                                elif ResourceCount == -2:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                elif ResourceCount == -3:
                                        ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                        ResourceCount = ResourceCount*2
                        elif LuxuryQuantity == -4:
                                ResourceCount = Players
                        elif LuxuryQuantity > -4:
                                ResourceCount = (Players + 1)/2 + dice.get(Players/2, "How many of the resource?")
                                if LuxuryQuantity == -1:
                                        ResourceCount = ResourceCount/2
                                elif LuxuryQuantity == -3:
                                        ResourceCount = ResourceCount*2

		# Remove eligible tiles that already contain a resource.

		for I in ResourceLocations:
                        if I in EligibleTiles[Resource]:
                                EligibleTiles[Resource].remove(I)

		# Place the resource.
		
		while (added < ResourceCount and tries < 4*ResourceCount):

                        tries += 1

			# Are there any available plots for the resource?
			
			if EligibleTiles[Resource] == []:
				break

			# Pick an eligible tile.

			EligibleTile = dice.get(len(EligibleTiles[Resource]), "A random eligible tile.")
			I = EligibleTiles[Resource][EligibleTile]
			del EligibleTiles[Resource][EligibleTile]

			NearResource = 1
			NearResources = 1

			for i in PlayerStartCross:
                                Inew = GetNewTileIndex(I, i)
                                if Inew in NearLocations:
                                        NearResource += 1
                                if Inew in ResourceLocations:
                                        NearResources += 1

                        RandOne = 1 + dice.get(NearResource, "Odds decrease with clumping of the resource.")
                        RandTwo = 3 + dice.get(NearResources, "Odds decrease with clumping of resources.")

                        if (RandOne >= NearResource and RandTwo >= NearResources):
                                Plot = cymap.plotByIndex(I)
                                Plot.setBonusType(Resource)
                                added += 1
                                ResourceLocations.append(I)
                                NearLocations.append(I)

                                # Ensure no illegal clumping of the resource takes place.

                                if not ResourceClumps[Resource]:
                                        for i in PreventResourceClumping:
                                                Inew = GetNewTileIndex(I, i)
                                                if Inew == -1:
                                                        continue
                                                if Inew in EligibleTiles[Resource]:
                                                        EligibleTiles[Resource].remove(Inew)

	return 0

#--------------------------------------------------------------------------------------------------------------------------#
# Step 6: Place huts.
#--------------------------------------------------------------------------------------------------------------------------#

def addGoodies():
        
	CyPythonMgr().allowDefaultImpl()
	return

#--------------------------------------------------------------------------------------------------------------------------#
# Step 7: Players are placed.
#--------------------------------------------------------------------------------------------------------------------------#

# The first called function will be used to generate the starting plots.
# The second called function will be use to add rivers.
# The third called function will be used to alter the fat crosses of each starting tile.
def assignStartingPlots():

        StartingGeneration = getSelectedMapValue("Starting Positions:")

        if StartingGeneration == 0:
                CyPythonMgr().allowDefaultImpl()
                return

        cymap = CyMap()
	gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()

        global Area
        global NewplotTypes
        global UpdatedOwnerLandMasses
        global LandMassTileCount
        global Details
        global AssignedStartTiles

        Players = gc.getGame().countCivPlayersEverAlive()
        PossibleStartTiles = []*len(LandMassTileCount)
        LandPiece = []
        LandPieceLandMass = []
        LandPieces = 0
        check = False

        # Determine what tiles make up each piece of land and which land mass they belong to.
        
        for I in range(Area):

                # Ignore water tiles.

                if NewplotTypes[I] == PlotTypes.PLOT_OCEAN:
                        continue

                # Ignore tiles that have already been assigned to a piece of land.

                for i in range(LandPieces):
                        if I in LandPiece[i]:
                                check = True
                                break
                
                if check:
                        check = False
                        continue

                LandPiece.append([])
                NewLandPiece = [I]

                # What other tiles are part of the piece of land.

                while len(NewLandPiece) > 0:

                        # Take the next tile that forms the piece of land.
                        
                        J = NewLandPiece[0]

                        # Determine which of its neighbours help form the piece of land.

                        for i in nearDirs:
                                K = GetNewTileIndex(J, i)
                                if NewplotTypes[K] == PlotTypes.PLOT_OCEAN:
                                        continue
                                if K in NewLandPiece:
                                        continue
                                if K in LandPiece[LandPieces]:
                                        continue
                                NewLandPiece.append(K)

                        NewLandPiece.remove(J)
                        LandPiece[LandPieces].append(J)

                LandPieces += 1
                LandPieceLandMass.append(UpdatedOwnerLandMasses[I])

        # If terra options have been selected determine which land masses will be initially uninhabited.

        TerraLands = getSelectedMapValue("Min. New Worlds:")
        MaxTerraLands = getSelectedMapValue("Max. New Worlds:")
        LandMasses = max(UpdatedOwnerLandMasses) + 1
        
        if (MaxTerraLands >= 0 and MaxTerraLands != TerraLands):
                DiffTerraLands = abs(MaxTerraLands - TerraLands) + 1
                TerraLands = min(TerraLands, MaxTerraLands) + dice.get(DiffTerraLands, "The number of new worlds.")

        if TerraLands >= LandMasses:
                TerraLands = LandMasses - 1

        NewWorlds = []

        for i in range(TerraLands):
                LandMass = dice.get(LandMasses, "Pick a land mass to be initially uninhabited.")
                while LandMass in NewWorlds:
                        LandMass = dice.get(LandMasses, "Pick a land mass to be initially uninhabited.")
                NewWorlds.append(LandMass)

        for Land in range(LandPieces):
                if LandPieceLandMass[Land] in NewWorlds:
                        LandPiece[Land] = []

        # Determine the size of each piece of land, excluding peaks, snow, tundra, and featureless desert.

        LandSize = [0]*LandPieces

        for Land in range(LandPieces):
                if LandPiece[Land] == []:
                        continue
                for I in LandPiece[Land]:
                        
                        if NewplotTypes[I] == PlotTypes.PLOT_PEAK:
                                LandPiece[Land].remove(I)
                                continue
                        
                        Plot = cymap.plotByIndex(I)
                        Terrain = Plot.getTerrainType()
                        Feature = Plot.getFeatureType()

                        if Terrain not in [0, 1]:
                                if Feature not in [2, 3]:
                                        LandPiece[Land].remove(I)
                                        continue

                LandSize[Land] = len(LandPiece[Land])

        # Determine how many civs will be assigned to each piece of land.

        LandPlayers = [0]*LandPieces
        PlayersLand = [0]*Players
        LandSizePlayersRatio = [0]*LandPieces

        for Player in range(Players):

                # Find the land piece with the largest ratio of land size to the number of assigned players.

                MaxRatio = 0
                MaxRatioLand = 0

                for Land in range(LandPieces):
                        LandSizePlayersRatio[Land] = 1.0*LandSize[Land]/(LandPlayers[Land] + 1)
                        if LandSizePlayersRatio[Land] > MaxRatio:
                                MaxRatio = LandSizePlayersRatio[Land]
                                MaxRatioLand = Land

                # Assign another player to start on this piece of land.

                LandPlayers[MaxRatioLand] += 1
                PlayersLand[Player] = MaxRatioLand

        # Find the starting locations for each civ.

        StartTiles = []
        check = False

        for Land in range(LandPieces):

                # If no one starts on this piece of land, go straight to the next piece of land.
                
                if LandPlayers[Land] <= 0:
                        continue

                # Determine whether a tile on the piece of land is coastal and how many water tiles are in its fat cross.

                WaterTiles = []
                PossibleStartTiles = []

                # Determine each tiles coastal status.

                for Tile in range(LandSize[Land]):
                        I = LandPiece[Land][Tile]

                        # First exclude all snow and tundra bound starts.

                        Plot = cymap.plotByIndex(I)
                        Terrain = Plot.getTerrainType()

                        if Terrain == 4 or Terrain == 3:
                                continue

                        # Exclude tiles that contain resources.

                        if Plot.getBonusType(-1) in [2, 3, 4, 9, 11, 12, 14, 16, 17, 18, 19, 22, 23, 27]:
                                continue

                        # Begin to determine the tiles coastal status.
                        
                        FatCrossWaterTiles = 0
                        for i in nearDirs:
                                J = GetNewTileIndex(I, i)
                                if NewplotTypes[J] == PlotTypes.PLOT_OCEAN:
                                        FatCrossWaterTiles += 1

                        # Exclude tiles that aren't coastal if only 1 civ starts on the piece of land.

                        if LandPlayers[Land] == 1 and FatCrossWaterTiles == 0:
                                continue

                        # If the tile is not coastal, exclude it if its fat cross contains water tiles.

                        elif FatCrossWaterTiles == 0:
                                for i in range(8, 20):
                                        J = GetNewTileIndex(I, PlayerStartCross[i])
                                        if NewplotTypes[J] == PlotTypes.PLOT_OCEAN:
                                                FatCrossWaterTiles += 1
                                                break
                                if FatCrossWaterTiles > 0:
                                        continue
                                else:
                                        FatCrossWaterTiles = 4

                        # If the tile is coastal, find how many water tiles are in its fat cross.

                        else:
                                for i in range(8, 20):
                                        J = GetNewTileIndex(I, PlayerStartCross[i])
                                        if NewplotTypes[J] == PlotTypes.PLOT_OCEAN:
                                                FatCrossWaterTiles += 1

                        # Collect the data for possible starting tiles.

                        PossibleStartTiles.append(I)
                        WaterTiles.append(FatCrossWaterTiles)

                PossibleStarts = len(PossibleStartTiles)

                # If only one civ is to be placed, place it so that:
                # 1) It is on a coastal tile,
                # 2) Its fat cross contains the least number of water tiles,
                # 3) It is as close to the center of the piece of land as possible.

                if LandPlayers[Land] == 1:

                        # Find a tile with the least number of water tiles.

                        MinWaterStarts = []
                        MinWater = min(WaterTiles)

                        for i in range(PossibleStarts):
                                if WaterTiles[i] == MinWater:
                                        MinWaterStarts.append(PossibleStartTiles[i])

                        # If there is more than one such tile, find the tile closest to the center of the land.

                        if len(MinWaterStarts) > 1:
                                MaxI = max(LandPiece[Land])
                                MinI = min(LandPiece[Land])
                                AverageI = (MaxI + MinI)/2
                                LeastDifference = Area
                                for I in MinWaterStarts:
                                        X, Y = GetDistance(I, AverageI)
                                        Dist = (X**2 + Y**2)**(0.5)
                                        if Dist < LeastDifference:
                                                LeastDifference = Dist
                                                MidTile = I
                                StartTiles.append(MidTile)
                        
                        # If only one coastal tile has a fat cross with the least number of water squares.

                        else:
                                StartTiles.append(MinWaterStarts[0])

                        continue

                # If there is more than one civ to be placed on this piece of land, use SevenSpirits placement code:
                # 1) Determine the value of a city placed on each possible start tile.
                # 2) Starting with the best tile, remove all tiles within 3 tiles of each tile.
                # 3) Slowly expand the area of elimination until there are only enough possible start tiles left.

                # Determine the value of each possible start tile.
                # Factors accounted for are coastal status, the number of coastal tiles, and latitude.

                OrderedPossibleStarts = []
                OrderedWaterTiles = []

                while len(OrderedPossibleStarts) < PossibleStarts:
                        OrderedPossibleStartLength = len(OrderedPossibleStarts)
                        MinWater = min(WaterTiles)
                        MinWaterTiles = []
                        ReorderingWaterTiles = []
                        TileLatitude = []
                        for j in range(PossibleStarts):
                                if WaterTiles[j] == MinWater:
                                        MinWaterTiles.append(PossibleStartTiles[j])
                                        ReorderingWaterTiles.append(WaterTiles[j])
                                        TileLatitude.append(abs(PossibleStartTiles[j] - Area/2))
                                        WaterTiles[j] = 21
                        while len(OrderedPossibleStarts) < OrderedPossibleStartLength + len(MinWaterTiles):
                                ClosestEquator = min(TileLatitude)
                                for j in range(len(MinWaterTiles)):
                                        if TileLatitude[j] == ClosestEquator:
                                                OrderedPossibleStarts.append(MinWaterTiles[j])
                                                OrderedWaterTiles.append(ReorderingWaterTiles[j])
                                                TileLatitude[j] = Area

                # In order of best tile to worst tile, eliminate all other tiles within 3 tiles of each selected tile.

                Eliminated = 0

                for i in range(PossibleStarts):
                        I = OrderedPossibleStarts[i]
                        if I == -1:
                                continue
                        for j in range(PossibleStarts):
                                if i == j:
                                        continue
                                J = OrderedPossibleStarts[j]
                                if J == -1:
                                        continue
                                X, Y = GetDistance(I, J)
                                if abs(X) < 3 and abs(Y) < 3:
                                        OrderedPossibleStarts[j] = -1
                                        OrderedWaterTiles[j] = -1
                                        Eliminated += 1
                
                PossibleStarts -= Eliminated

                while min(OrderedPossibleStarts) == -1:
                        OrderedPossibleStarts.remove(-1)

                while min(OrderedWaterTiles) == -1:
                        OrderedWaterTiles.remove(-1)

                # Slowly expand the radius of elimination until just enough tiles are left.

                ViableStartTiles = range(PossibleStarts)
                ReversedTileIndicies = range(PossibleStarts)
                ReversedTileIndicies.reverse()
                EliminatedTiles = 0
                Dist = -3
                check = False

                while check == False:
                        Dist += 1
                        for i in range(PossibleStarts):
                                if check:
                                        break
                                if ViableStartTiles[i] == -1:
                                        continue
                                I = OrderedPossibleStarts[i]
                                for j in ReversedTileIndicies:
                                        if check:
                                                break
                                        if ViableStartTiles[j] == -1:
                                                continue
                                        if i == j:
                                                continue
                                        J = OrderedPossibleStarts[j]
                                        X, Y = GetDistance(I, J)
                                        PathDistance = max([abs(X), abs(Y)]) - OrderedWaterTiles[j]/4
                                        if PathDistance <= Dist:
                                                if ViableStartTiles[j] == j:
                                                        ViableStartTiles[j] = i
                                                else:
                                                        EliminatedTiles += 1
                                                        ViableStartTiles[j] = -1
                                                        if EliminatedTiles + LandPlayers[Land] == PossibleStarts:
                                                                check = True

                for i in range(PossibleStarts):
                        if ViableStartTiles[i] > -1:
                                StartTiles.append(OrderedPossibleStarts[i])

        # Place each civ.

        AssignedStartTiles = []

        for Player in range(Players):
                StartTileIndex = dice.get(len(StartTiles), "Pick a tile for a civ to start on.")
                StartTile = cymap.plotByIndex(StartTiles[StartTileIndex])
                AssignedStartTiles.append(StartTiles[StartTileIndex])
                del StartTiles[StartTileIndex]
                StartTile.setStartingPlot(True)
                gc.getPlayer(Player).setStartingPlot(StartTile, True)

        return

def normalizeAddRiver():
        CyPythonMgr().allowDefaultImpl()
        return

def normalizeRemovePeaks():

        # In order to ensure that all appropriate tiles are flood plains;

        global Area
        global NewplotTypes

        cymap = CyMap()
        gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()

        FloodPlains = getSelectedMapValue("Flood Plains:")

        if FloodPlains == 1:
                for I in range(Area):
                        if NewplotTypes[I] != PlotTypes.PLOT_LAND:
                                continue
                        Plot = cymap.plotByIndex(I)
                        if not Plot.isRiverSide():
                                continue
                        Terrain = Plot.getTerrainType()
                        Feature = Plot.getFeatureType()
                        if Terrain == 2 and Feature < 0:
                                Plot.setFeatureType(3, -1)

        elif FloodPlains == 2:
                for I in range(Area):
                        if NewplotTypes[I] != PlotTypes.PLOT_LAND:
                                continue
                        Plot = cymap.plotByIndex
                        if not Plot.isRiverSide():
                                continue
                        Terrain = Plot.getTerrainType()
                        Feature = Plot.getFeatureType()
                        if Terrain == 2 and Feature < 0:
                                Plot.setFeatureType(3, -1)
                        elif Terrain in [0, 1, 3] and Feature < 0:
                                if dice.get(3, "Place a flood plains") == FloodPlains:
                                        Plot.setFeatureType(3, -1)

        # Replace peaks in the fat crosses of any start tiles with hills.

        CyPythonMgr().allowDefaultImpl()
        return

def normalizeAddLakes():
        CyPythonMgr().allowDefaultImpl()
        return

def normalizeRemoveBadFeatures():
        CyPythonMgr().allowDefaultImpl()
        return

def normalizeRemoveBadTerrain():
        CyPythonMgr().allowDefaultImpl()
        return

def normalizeAddFoodBonuses():
        CyPythonMgr().allowDefaultImpl()
        return

def normalizeAddGoodTerrain():
        CyPythonMgr().allowDefaultImpl()
        return

def normalizeAddExtras():

        ForestSpam = getSelectedMapValue("Skip BTS Forests:")

        if ForestSpam == 0:
                CyPythonMgr().allowDefaultImpl()
                return

        cymap = CyMap()
        gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()

        global Area
        global width
        global AssignedStartingPlots

        IgnoreLatitude = False
        TotalValue = 0
	PlayerCount = 0
	BestValue = 0
	WorstValue = Area

	#for i in range(len(AssignedStartingPlots)):
  #              I = AssignedStartingPlots[i]
   #             X, Y = GetXY(I)
    #            Value = getPlayer(i).AI_foundValue(X, Y, -1, True)
     #           TotalValue += Value
      #          BestValue = max(TotalValue, BestValue)
       #         WorstValue = min(TotalValue, WorstValue)

        #TargetValue = BestValue*4/5

        #for i in range(len(AssignedStartingPlots)):
         #       I = AssignedStartingPlots[i]
          #      X, Y = GetXY(I)
           #     Count = 0
            #    FeatureCount = 0
             #   for J in AssignedStartingPlots:
              #          if getPlayer(i).AI_foundValue(X, Y, -1, True) >= TargetValue:
               #                 break
                        
                
        #return
#--------------------------------------------------------------------------------------------------------------------------#
# Step 8: Add additional menu options.
#--------------------------------------------------------------------------------------------------------------------------#

selection_titles = [unicode("Min Width:"),
                    unicode("Max Width:"),
                    unicode("Min Height:"),
                    unicode("Max Height:"),
		    unicode("Min Land Masses:"),
                    unicode("Max Land Masses:"),
                    unicode("Min Continents:"),
                    unicode("Max Continents:"),
                    unicode("Continent Distribution:"),
                    unicode("Continent Placement:"),
                    unicode("Min Oceanic Seperation:"),
                    unicode("Max Oceanic Seperation:"),
                    unicode("Min Ocean:"),
                    unicode("Max Ocean:"),
                    unicode("Coastlines:"),
                    unicode("Hills:"),
                    unicode("Peaks:"),
                    unicode("Mountain Ranges:"),
                    unicode("Altitude Changes:"),
                    unicode("Climate:"),
                    unicode("Desert:"),
                    unicode("Plains:"),
                    unicode("Terrain Clumping:"),
                    unicode("Jungle:"),
                    unicode("Forest:"),
                    unicode("Flood Plains:"),
                    unicode("Resources:"),
                    unicode("Resource Land Masses:"),
                    unicode("How Many Land Masses:"),
                    unicode("Resource Tiles:"),
                    unicode("Strategic Supply:"),
                    unicode("Health Supply:"),
                    unicode("Luxury Supply:"),
                    unicode("Let Resources Clump:"),
                    unicode("Starting Positions:"),
                    unicode("Min. New Worlds:"),
                    unicode("Max. New Worlds:"),
                    unicode("Skip BTS Forests:"),
                    unicode("Print Map Details:"),
                    ]

selection_names_and_values = [
        [
                ["16", 4],
                ["24", 6],
                ["32", 8],
                ["40 - Continents Duel", 10],
                ["48", 12],
                ["52 - Continents Tiny", 13],
                ["56", 14],
                ["64 - Continents Small", 16],
                ["72", 18],
                ["80", 20],
                ["84 - Continents Standard", 21],
                ["88", 22],
                ["96", 24],
                ["104 - Continents Large", 26],
                ["112", 28],
                ["120", 30],
                ["128 - Continents Huge", 32],
                ["136", 34],
                ["144", 36],
                ["152", 38],
                ["160", 40],
                ["168", 42],
                ["176", 44],
                ["184", 46],
                ["192", 48],
                ["200", 50],
                ["208", 52],
                ["216", 54],
                ["224", 56],
                ["232", 58],
                ["240", 60],
                ["248", 62],
                ["256", 64],
                ["264", 66],
                ["272", 68],
                ["280", 70],
                ["288", 72],
                ["296", 74],
                ["304", 76],
                ],
        [
                ["No variation", 0],
                ["16", 4],
                ["24", 6],
                ["32", 8],
                ["40 - Continents Duel", 10],
                ["48", 12],
                ["52 - Continents Tiny", 13],
                ["56", 14],
                ["64 - Continents Small", 16],
                ["72", 18],
                ["80", 20],
                ["84 - Continents Standard", 21],
                ["88", 22],
                ["96", 24],
                ["104 - Continents Large", 26],
                ["112", 28],
                ["120", 30],
                ["128 - Continents Huge", 32],
                ["136", 34],
                ["144", 36],
                ["152", 38],
                ["160", 40],
                ["168", 42],
                ["176", 44],
                ["184", 46],
                ["192", 48],
                ["200", 50],
                ["208", 52],
                ["216", 54],
                ["224", 56],
                ["232", 58],
                ["240", 60],
                ["248", 62],
                ["256", 64],
                ["264", 66],
                ["272", 68],
                ["280", 70],
                ["288", 72],
                ["296", 74],
                ["304", 76],
                ],
        [
                ["Normal width/height ratio", 0],
                ["16", 4],
                ["24 - Continents Duel", 6],
                ["32 - Continents Tiny", 8],
                ["40 - Continents Small", 10],
                ["48", 12],
                ["52 - Continents Standard", 13],
                ["56", 14],
                ["64 - Continents Large", 16],
                ["72", 18],
                ["80 - Continents Huge", 20],
                ["84", 21],
                ["88", 22],
                ["96", 24],
                ["104", 26],
                ["112", 28],
                ["120", 30],
                ["128", 32],
                ["136", 34],
                ["144", 36],
                ["152", 38],
                ["160", 40],
                ["168", 42],
                ["176", 44],
                ["184", 46],
                ["192", 48],
                ["200", 50],
                ["208", 52],
                ["216", 54],
                ["224", 56],
                ["232", 58],
                ["240", 60],
                ],
        [
                ["No variation", 0],
                ["16", 4],
                ["24 - Continents Duel", 6],
                ["32 - Continents Tiny", 8],
                ["40 - Continents Small", 10],
                ["48", 12],
                ["52 - Continents Standard", 13],
                ["56", 14],
                ["64 - Continents Large", 16],
                ["72", 18],
                ["80 - Continents Huge", 20],
                ["84", 21],
                ["88", 22],
                ["96", 24],
                ["104", 26],
                ["112", 28],
                ["120", 30],
                ["128", 32],
                ["136", 34],
                ["144", 36],
                ["152", 38],
                ["160", 40],
                ["168", 42],
                ["176", 44],
                ["184", 46],
                ["192", 48],
                ["200", 50],
                ["208", 52],
                ["216", 54],
                ["224", 56],
                ["232", 58],
                ["240", 60],
                ],
        [
                ["1 land mass", 1],
                ["2 land masses", 2],
                ["3 land masses", 3],
                ["4 land masses", 4],
                ["5 land masses", 5],
                ["6 land masses", 6],
                ["7 land masses", 7],
                ["8 land masses", 8],
                ["9 land masses", 9],
                ["10 land masses", 10],
                ],
        [
                ["No variation", 0],
                ["1 land mass", 1],
                ["2 land masses", 2],
                ["3 land masses", 3],
                ["4 land masses", 4],
                ["5 land masses", 5],
                ["6 land masses", 6],
                ["7 land masses", 7],
                ["8 land masses", 8],
                ["9 land masses", 9],
                ["10 land masses", 10],
                ],
        [
                ["1 continent per land mass",0],
                ["1 continent",1],
                ["2 continents",2],
                ["3 continents",3],
                ["4 continents",4],
                ["5 continents",5],
                ["6 continents",6],
                ["7 continents",7],
                ["8 continents",8],
                ["9 continents",9],
                ["10 continents",10],
                ["11 continents",11],
                ["12 continents",12],
                ["13 continents",13],
                ["14 continents",14],
                ["15 continents",15],
                ["16 continents",16],
                ["17 continents",17],
                ["18 continents",18],
                ["19 continents",19],
                ["20 continents",20],
                ],
        [
                ["No variation",0],
                ["1 continent",1],
                ["2 continents",2],
                ["3 continents",3],
                ["4 continents",4],
                ["5 continents",5],
                ["6 continents",6],
                ["7 continents",7],
                ["8 continents",8],
                ["9 continents",9],
                ["10 continents",10],
                ["11 continents",11],
                ["12 continents",12],
                ["13 continents",13],
                ["14 continents",14],
                ["15 continents",15],
                ["16 continents",16],
                ["17 continents",17],
                ["18 continents",18],
                ["19 continents",19],
                ["20 continents",20],
                ],
        [
                ["Evenly distributed continents", True],
                ["Randomly distributed continents", False],
                ["Random", -1],
                ],
        [
                ["Normal continent placement", 0],
                ["More rounded land masses", 1],
                ["More snaky land masses", 2],
                ["Random", -1],
                ],
        [
                ["No separation",0],
                ["1 tile of seperation",1],
                ["2 tiles of seperation",2],
                ["3 tiles of seperation",3],
                ["4 tiles of seperation",4],
                ["5 tiles of seperation",5],
                ["6 tiles of seperation",6],
                ["No overseas trade before astronomy",7],
                ["8 tiles of seperation",8],
                ["9 tiles of seperation",9],
                ["10 tiles of seperation",10],
                ["11 tiles of seperation",11],
                ["12 tiles of seperation",12],
                ],
        [
                ["No variation",-1],
                ["No separation",0],
                ["1 tile of separation",1],
                ["2 tiles of seperation",2],
                ["3 tiles of seperation",3],
                ["4 tiles of seperation",4],
                ["5 tiles of seperation",5],
                ["6 tiles of seperation",6],
                ["No overseas trade before astronomy",7],
                ["8 tiles of seperation",8],
                ["9 tiles of seperation",9],
                ["10 tiles of seperation",10],
                ["11 tiles of seperation",11],
                ["12 tiles of seperation",12],
                ],
        [
                ["5% ocean",95],
                ["10% ocean",90],
                ["15% ocean",85],
                ["20% ocean",80],
                ["25% ocean",75],
                ["30% ocean",70],
                ["35% ocean",65],
                ["40% ocean",60],
                ["45% ocean",55],
                ["50% ocean",50],
                ["55% ocean",45],
                ["60% ocean",40],
                ["65% ocean",35],
                ["70% ocean",30],
                ["75% ocean",25],
                ["80% ocean",20],
                ["85% ocean",15],
                ["90% ocean",10],
                ["95% ocean",5],
                ],
        [
                ["No variation",0],
                ["5% ocean",95],
                ["10% ocean",90],
                ["15% ocean",85],
                ["20% ocean",80],
                ["25% ocean",75],
                ["30% ocean",70],
                ["35% ocean",65],
                ["40% ocean",60],
                ["45% ocean",55],
                ["50% ocean",50],
                ["55% ocean",45],
                ["60% ocean",40],
                ["65% ocean",35],
                ["70% ocean",30],
                ["75% ocean",25],
                ["80% ocean",20],
                ["85% ocean",15],
                ["90% ocean",10],
                ["95% ocean",5],
                ],
        [
                ["Standard", 2],
                ["Rougher", 3],
                ["Rugged", 4],
                ["Random", -1],
                ],
        [
                ["Few", -1],
                ["Some", -2],
                ["Normal", -3],
                ["Many", -4],
                ["Many many", -5],
                ["Many many many", -6],
                ["Many many many many", -7],
                ["Lots", -8],
                ["Random", -9],
                ["Highlands few", 40],
                ["Highlands normal", 45],
                ["Highlands many", 50],
                ["Random highlands", 0],
                ],
        [
                ["Few", -1],
                ["Some", -2],
                ["Normal", -3],
                ["Many", -4],
                ["Many many", -5],
                ["Many many many", -6],
                ["Many many many many", -7],
                ["Lots", -8],
                ["Random", -9],
                ["Highlands few", 20],
                ["Highlands normal", 25],
                ["Highlands many", 30],
                ["Random highlands", 0],
                ],
        [
                ["No ranges", 0],
                ["Ranges", 1],
                ["All ranges", 2],
                ["Random", -1],
		],
        [
                ["No clumped hills and peaks", 0],
                ["Clumped hills and peaks", 1],
                ["Random", -1],
		],
        [
                ["Warm", 0],
                ["Temperate", 1],
                ["Cool", 2],
                ["Random", -1],
		],
        [
                ["Less desert", 16],
                ["Normal desert", 32],
                ["More desert", 48],
                ["Random", 1],
		],
        [
                ["Less plains", 9],
                ["Normal plains", 18],
                ["More plains", 27],
                ["Random", 1],
		],
        [
                ["Very clumped terrain", 2],
                ["Clumped terrain", 3],
                ["Normal", 4],
                ["Varied terrain", 5],
                ["Very varied terrain", 6],
                ["Random", 1],
                ],
        [
                ["Less jungle", 100],
                ["Normal jungle", 80],
                ["More jungle", 60],
                ["Random", 1],
		],
        [
                ["Less forest", 75],
                ["Normal forest", 60],
                ["More forest", 45],
                ["Random", 1],
		],
        [
                ["Normal flood plains placement", 1],
                ["Extra flood plains", 2],
                ],
        [
                ["Standard - ignore resource options", False],
                ["Custom - activate resource options", True],
                ],
        [
                ["Standard", 0],
                ["LandMasses", 1],
                ["LandMasses, isolate horses", 2],
                ["No isolated resources", 3],
                ["Random", -1],
                ["Custom", 4],
                ],
        [
                ["Assigned to 1 land mass", 1],
                ["Assigned to 2 land masses", 2],
                ["Assigned to 3 land masses", 3],
                ["Assigned to 4 land masses", 4],
                ["Assigned to 5 land masses", 5],
                ["Assigned to 6 land masses", 6],
                ["Assigned to 7 land masses", 7],
                ["Assigned to 8 land masses", 8],
                ["Random", 0],
                ["Assigned to half the land masses", -1],
                ["Assigned to all but 1 land mass", -2],
                ["Custom", -3],
                ],
        [
                ["Standard", 0],
                ["LandMasses", 1],
                ["Crazy", 2],
                ["Custom", 3],
                ],
        [
                ["Few strategic resources", -1],
                ["Normal strategic resources", -2],
                ["Many strategic resources", -3],
                ["Random", 0],
                ["1 of each per civilization", -4],
                ["Custom", -5],
                ],
        [
                ["Few health resources", -1],
                ["Normal health resources", -2],
                ["Many health resources", -3],
                ["Random", 0],
                ["1 of each per civilization", -4],
                ["Custom", -5],
                ],
        [
                ["Few luxury resources", -1],
                ["Normal luxury resources", -2],
                ["Many luxury resources", -3],
                ["Random", 0],
                ["1 of each per civilization", -4],
                ["Custom", -5],
                ],
        [
                ["Standard", [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
                ["None", [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],
                ["Strategic", [1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],
                ["Health", [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],
                ["Luxury", [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
                ["Strategic/Health", [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],
                ["Strategic/Luxury", [1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
                ["Health/Luxury", [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
                ["Strategic/Health/Luxury", [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]],
                ["Custom", CustomiseResourceClumps()],
                ],
        [
                ["Standard placement", 0],
                ["LandMasses placement", 1],
                ],
        [
                ["No uninhabited land masses", 0],
                ["1 uninhabited land mass", 1],
                ["2 uninhabited land masses", 2],
                ["3 uninhabited land masses", 3],
                ["4 uninhabited land masses", 4],
                ["5 uninhabited land masses", 5],
                ["6 uninhabited land masses", 6],
                ["7 uninhabited land masses", 7],
                ["8 uninhabited land masses", 8],
                ["9 uninhabited land masses", 9],
                ],
        [
                ["No variation", -1],
                ["No uninhabited land masses", 0],
                ["1 uninhabited land mass", 1],
                ["2 uninhabited land masses", 2],
                ["3 uninhabited land masses", 3],
                ["4 uninhabited land masses", 4],
                ["5 uninhabited land masses", 5],
                ["6 uninhabited land masses", 6],
                ["7 uninhabited land masses", 7],
                ["8 uninhabited land masses", 8],
                ["9 uninhabited land masses", 9],
                ],
        [
                ["No", 0],
                ["Yes", 1],
                ],
        [
                ["No", False],
                ["Yes", True],
                ],
        ]

def isSeaLevelMap():
        return 0

def isRandomCustomMapOption(argsList):
        return False

def getDescription():
	return unicode("LandMasses")

def getNumCustomMapOptions():
	return len(selection_names_and_values)

def getCustomMapOptionName(argsList):
	index = argsList[0]
	return selection_titles[index]

def getSelectedMapValue(optionName):
        
	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()
	
	for optionIndex in range(len(selection_titles)):
		title = selection_titles[optionIndex]
		if title == optionName:
			break
	
	userSelectedIndex = int(cymap.getCustomMapOption(optionIndex))
	
	if userSelectedIndex >= len(selection_names_and_values[optionIndex]):
		userSelectedIndex = 0
				
	cachedMenuChoices[optionIndex] = userSelectedIndex
	
	return selection_names_and_values[optionIndex][userSelectedIndex][1]

def getNumCustomMapOptionValues(argsList):
	index = argsList[0]
	return len(selection_names_and_values[index])

def getCustomMapOptionDescAt(argsList):
	iOption = argsList[0]
	iSelection = argsList[1]
	return unicode(selection_names_and_values[iOption][iSelection][0])

# Save selected map options.

def saveSettings():
	cymap = CyMap()
	settings = open('LandMasses.cfg', 'w')
	smoptions = []
	index = -1
	for smoption in selection_titles:
		index += 1
		selectionIndex = int(cymap.getCustomMapOption(index))
		valueName = selection_names_and_values[index][selectionIndex][0]
		smoptions.append((smoption,valueName))
		
	print smoptions
	try:
		pickle.dump(smoptions, settings)
	except Exception, inst:
		print "    Pickling Error trying to save smartmap settings to LandMasses.cfg", inst
	settings.close()

# Find previously selected map options.

def getCustomMapOptionDefault(argsList):
	index = argsList[0]
	result = 0
	try:
		settings = open('LandMasses.cfg', 'r')
		optionsWithDefault = pickle.load(settings)
		for optionIndex in range(len(optionsWithDefault)):
			option, valueName = optionsWithDefault[index]
			if option == selection_titles[index]:
				for selectionIndex in range(len(selection_names_and_values[index])):
					selectionName, selectionValue = selection_names_and_values[index][selectionIndex]
					if selectionName == valueName:
						result = selectionIndex
		settings.close()
	except IOError:
		print "    Couldn't find LandMasses.cfg"
	except EOFError:
		print "    Bad contents in LandMasses.cfg"
		
	return result

def isAdvancedMap():
	return 1

#--------------------------------------------------------------------------------------------------------------------------#
# Step 9: Further functions.
#--------------------------------------------------------------------------------------------------------------------------#

def beforeInit():
	global cachedMenuChoices
	cachedMenuChoices = []
	for i in range(len(selection_names_and_values)):
		cachedMenuChoices.append((0))
		
def beforeGeneration():
	global cachedMenuChoices
	print "    before generation"
	for i in range(len(selection_names_and_values)):
		if selection_names_and_values[i][0] != "Wrap:":
			cachedMenuChoices[i] = 0

def GetNewTileIndex(I, direction):
	
	global width
	global Area

	dx, dy = direction
	Inew = I + dx + (width*dy)

	if Inew < 0 or Inew >= Area:
		Inew = -1

	return (Inew)

def GetDistance(Ione, Itwo):
	
	global width

	x = abs(Ione - Itwo)
	y = 0

	while x >= width:
		x -= width
		y += 1

	if Itwo > Ione:
		y = -y

	if x > int(width/2):
		x = width - x
			
	return (x, y)

def GetXY(I):

        global width

        X = I
        Y = 0

        while X >= width:
                X -= width
                Y += 1

        return (X, Y)

def GetLandMasses():

        global Area

        cymap = CyMap()
	
	LandMassTiles = []
	LandMasses = []
	LandMassNumber = -1

	# Determine which tiles land masses and coasts are comprised of.

        for I in range(Area):

                Plot = cymap.plotByIndex(I)
                Terrain = Plot.getTerrainType()

                if Terrain == 6:
                        continue

                LandMassTiles.append(I)

        while len(LandMassTiles) > 0:

                # Pick the first tile of a landmass.

                LandMasses.append([])
                LandMassNumber += 1
                I = LandMassTiles[0]
                LandMassTiles.remove(I)
                LandMasses[LandMassNumber].append(I)
                AdjacentLandMassTiles = []

                # Next, collect its neighbours that also form the land mass.

                for i in cardinalDirs:

                        Inew = GetNewTileIndex(I, i)
                        Plotnew = cymap.plotByIndex(Inew)
                        Terrainnew = Plotnew.getTerrainType()

                        if Terrainnew == 6:
                                continue

                        AdjacentLandMassTiles.append(Inew)

                # Collect all tiles which form the land mass.

                while len(AdjacentLandMassTiles) > 0:

                        Inew = AdjacentLandMassTiles[0]
                        AdjacentLandMassTiles.remove(Inew)
                        LandMassTiles.remove(Inew)
                        LandMasses[LandMassNumber].append(Inew)

                        for i in cardinalDirs:

                                Inext = GetNewTileIndex(Inew, i)

                                if Inext in LandMasses[LandMassNumber] or Inext in AdjacentLandMassTiles:
                                        continue
                                
                                Plotnext = cymap.plotByIndex(Inext)
                                Terrainnext = Plotnext.getTerrainType()

                                if Terrainnext == 6:
                                        continue

                                AdjacentLandMassTiles.append(Inext)

        return LandMasses

#--------------------------------------------------------------------------------------------------------------------------#
# Version history.
#--------------------------------------------------------------------------------------------------------------------------#
# 1.0:
# Original version
#--------------------------------------------------------------------------------------------------------------------------#
# 1.1:
# Removed variables which weren't doing anything unique. This has resulted in slowdown in map generation - particularly
# for larger maps.
# Adjusted the percentage of land tiles that are hills and peaks.
# Increased the amount of hills and peaks possible through more options.
# fixed an error in the lake removal code which caused some lakes to be left behind. Note that lakes are added back in
# later.
# Fixed an error in the 'reduced desert' terrain option.
#--------------------------------------------------------------------------------------------------------------------------#
# 1.2:
# New way of choosing random numbers of land masses and continents implemented to allow greater flexibility.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.0:
# Alternative resource generator.
# - Oil can be generated anywhere.
# - Aquatic resources can be generated anywhere within reach of land.
# - Minerals (aluminium, coal, copper, gems, gold, iron, marble, silver, stone, uranium) and incense can be generated on
# any land tiles (except peaks and oases).
# - Other resources can be generated on the following terrain types.
# - Ice: fur.
# - Tundra: game, fur, ivory, sheep.
# - Tundra hills: game, fur, sheep.
# - Tundra forests: game, fur, ivory.
# - Tundra forested hills: game, fur.
# - Grassland: corn, cow, dye, horse, ivory, pig, rice, sheep, spices, sugar, wheat, wine.
# - Grassland hills: dye, pig, sheep, spices, wine.
# - Grassland forest or jungle: dye, fruit, ivory, pig, silk, spices, sugar, wine.
# - Grassland forested hills or jungle hills: dye, fruit, pig, silk, spices, sugar, wine.
# - Plains: Same as grassland but may also yield game and fur.
# Some resources are unique to their landmasses. These are corn, cow, pig, rice, sheep, wheat, ivory, silk, sugar, and wine.
# The option 'LandMasses: isolate horses' will include horses in this list.
# The option 'LandMasses: no isolated resources' will allow all resources to be placed on multiple land masses.
# LandMasses resource generator may override a mod which alters the placement of the standard resources.
# When using a mod with extra resources (salt, say) the LandMasses resource generator should (hopefully) place the
# additional resources according to the mods standard placement rules for that resource.
# Note that the options for isolating resources - 'LandMasses' and 'LandMasses: isolate horses' - do not completely isolate
# all resources selected for isolation. This is most prominently, but not completely, due to resources being placed within
# a civs starting place.
# Resources are placed randomly taking no account of the local resource density. This can lead to some serious clustering.
#
# Other changes:
# Increased the number of possible land masses to 10.
# Increase the possible width and height to 240 tiles (from 160).
# Increased the number of ocean percentages that can be chosen. The minimum is now 5% (was 40%) and the maximum is now 95%
# (was 80%).
# Added min. and max. options (previously included in the land masses and continents options) to the following options:
# - Width
# - Height
# - Ocean
# - Land Mass Seperation
# One affect of this change is that the 'Sealevel' option no longer has any effect on map generation. To obtain its affects:
# - Low: set Land Mass Seperation to between 0 and 4.
# - Medium: set Land Mass Seperation to between 5 and 8.
# - High: set Land Mass Seperation to between 9 and 12.
# Note the setting Land Mass Seperation at 0 or 1 will still cause only 1 continent per landmass to be generated.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.1:
#
# Terrain changes:
#
# Selecting Rocky climate no longer affects the terrain.
# Introduced three new hill and peaks setting; Highlands few, Highlands normal, and Highlands many. I've only used
# estimates to determine the appropriate percentages in these cases. If someone has the actual figures I'd appreciate them.
#
# Resource changes:
#
# Created a master resource option, 'Resources'.
# - Standard will use the standard CivIV recourse placement code. The Resource Land Mass, Resource Tiles, and
# Resource Quantities options are rendered ineffective.
# - Custom activates the Resource Land Mass, Resource Tiles, and Resource Quantities options.
#
# Divided resources options into two:
# - Resource Land Mass
# - Resource Tiles
# Resource Land Mass determines which resources are assigned to specific land masses.
# - 'Standard' assigns health and luxury resources to specific land masses.
# - 'LandMasses' assigns the same resources to specific land masses as it did previously.
# - 'LandMasses, isolate horses' is the same as 'LandMasses' but it also assigns horses to a specific land mass.
# - 'No isolated resources' allows all resources to be placed on all land masses.
# Resource Tiles determines which tiles a resource can be assigned to.
# - 'Standard' assigns resources to normal CivIV locations.
# - 'LandMasses' assigns the same resources to the same tile types as it did previously.
# - 'Crazy' assigns a resource to any tile which can be improved to take advantage of the resource.
# Setting both options to 'Standard' will cause LandMasses to use the default resource placement code.
# Setting only one option to 'Standard' will cause LandMasses to attempt to mimic the effects of the default resource
# placement code.
#
# Added the option 'Resource Quantities'. This allows the number of resources to be altered.
# - 'Normal resources': the no. of resources is between half the no. of players and one short of the no. of players.
# - 'Few resources': Only half as much of each resource are generated.
# - 'Many resources': Twice as much of each resources are generated.
# Further notes on this option:
# - if 'Standard' has been set for both Resource Land Mass and Resource Tiles then Resource Quantities is ignored.
# - The script may not place all of the resources that it sets out to.
# - The fewer resources generated, the more likely that the civ starting placement code may add in resources. This will
# noticably reduce the effects of some of the choices for Resource Land Mass, 'LandMasses', 'LandMasses, isolate horses',
# and, if 'Standard' has not been selected for Resource Tiles, 'Standard'.
#
# The odds of a resource being placed on any given tile reduce as the number of that resource in the fat cross centred
# around that tile increase.
# The odds of a resource being placed on any given tile reduce as the number of all resources in the fat cross centred
# around that tile increase.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.2:
#
# Beginning work on terrain options. Since my previous undocumented attempt at this (a few months ago) was a spectacular
# failure, I'll be going very slowly. For now only a few, unlikely-to-be-used, options are included.
#
# Terrain:
# - Standard: uses normal terrain generation. Deactivates all terrain options.
# - Custom: Activates all terrain options.
# Terrain Type:
# - all grass, all plains, all desert, all tundra, and all snow.
# - fantasy - randomly placed terrrain tiles.
#
# Resources:
# Aquatic resources will no longer be generated beneath ice.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.3:
#
# Lands:
#
# Added 'Coastlines' option. A rugged coastline increses the number of islands and archipelagos.
# - Standard: Prior to the introduction of the coastlines option, this was the usual implementation.
# - Rugged: Lots of coastal islands. Lands that were narrow under the standard setting may be broken up into archipelagos.
# Notes:
# - the effectiveness of this option depends a great deal on how much space is designated towards land, ocean,
# and seperation between landmasses. Selecting a low ocean percentage and a high seperation between land masses will
# reduce the effects of the coastline option to irrelevance.
# - it is possible for large pieces of land which previously formed a single landmass to be seperated from each other. If
# these lands become seperated by ocean (and not just coastal tiles) then they should be treated as seperate land masses
# by the resource generator. In short, you may get more land masses than you intended. The odds of this happening are much
# more likely with coastlines set to rugged but still possible (I've never seen it) under standard coatlines.
#
# The Ocean and Distance between land masses have been swapped around on the custom menu screen. This is so that the
# options related to the ratio of ocean tiles to land tiles are placed in order of importance:
# 1) Seperation Between Land Masses - land will only be placed if it doesn't go to close to other land masses. If the
# minimum seperation is sufficiently large enough there may not be enough space to reach the ocean percentage. For
# example, if the minimum seperation between land masses is set to 12 and the ocean percentage is set to 20%, the code
# will stop placing land tiles long before 80% of the map has been assigned a land tile. This does not include Pangaeas
# since the land mass seperation only affects the placement of tiles from different land masses.
# 2) Ocean: If the land mass seperation leaves enough space for the ocean/land ratio to be met then it should happen.
# 3) Coasts: The ruggedness of the coast depends on how much extra space there is for land after the ocean/land ratio set
# by the ocean option has been met. If there is no more space the land masses will be smooth. Selecting rugged will only
# have a noticeable affect beyond that of standard if there is plenty of room for extra land.
#
# Resources:
#
# Added options to give control over the quantities of strategic, health, and luxury resources. Ivory, spices, sugar,
# and wine are classified as luxuries. 
# Improved land mass placement of isolated resources. Resources selected for isolation are far less likely to appear on
# multiple land masses. The placement of resources on multiple land masses due to the civ placement code is unaffected.
# Aquatic resources should no longer appear on lakes.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.4:
#
# Lands:
#
# Added a third choice to the Coastlines option. Rougher' produces more rugged coasts than standard but smoother coasts
# than rugged. Like the other two choices available, the effectiveness of the option is based on the choices selected
# for Land Mass Seperation and Ocean.
#
# Resources:
#
# Added 'custom' choice to all resource options which allows a Python user to much more easily edit how resources are
# placed. The code which needs to be edited can be found under the heading CUSTOMISING RESOURCE PLACEMENT at the top of
# the LandMasses.py map script. Instructions can be found there explaining how to edit the resource placement.
# Custom 'Resource LandMasses' allows the user to select the resources which will be limited to one continent. By default,
# custom produces the same results as LandMasses.
# Custom 'Resource Tiles' allows the user to select the tiles a resource can be placed on. By default, custom produces
# the same results as LandMasses.
# Custom 'Strategic Quantities' allows the user to select the number of each strategic resource. By default, custom
# produces the same results as normal strategic resources.
# Custom 'Health Resources' and custom 'Luxury Resources' options work in the same way as custom 'Strategic Resources.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.5:
#
# Resources:
#
# Customising 'ResourceTiles' now allows the user to distinguish between coastal and ocean tiles.
# Customising 'Resource Land Masses' now allows water based resources to be assigned to particular land masses. Previously,
# customising the options so that water based resources were isolated to single land masses was ineffective.
# Introduced two new options: 'How many Land Masses' and 'Let Resources Clump'.
#
# 'How Many Land Masses':
# If a resource is selected to be isolated - not placed on all land masses - this options allows the user to determine
# how many the resource will be placed on.
# - Assigned to x land masses: Isolated resources are placed on x of the land masses. If x exceeds the number of land
# masses generated then no resources will be isolated; all resources selected for isolation may be generated on all land
# masses. Note that x only goes as far as 8. An alternative choice (listed below) exists if the user wants to place a
# resource on 9 out of 10 land masses.
# - Assigned to half the land masses: Isolated resources are placed on half of the generated land masses.
# - Assigned to all but one land mass: But for one land mass, an isolated resource may be placed anywhere.
# - Custom: Like all resource options, the ability to customise the effects of this option are available to users willing
# to edit this script with Python. Custom allows different resources to be isolated to different numbers of land
# masses (though the user must also make sure to set the appropriate resources to be isolated in the first place). Left
# unedited, this choice will assign isolated resources to just one land mass.
# Note that the map script is not guaranteed to generate the predefined number of landmasses so setting this option
# to 'assigned to x land masses' where x is a larger number may simply allow isolated resource to be generated everywhere.
#
# 'Let Resources Clump':
# This determines whether or not two of any one resource are allowed to be placed near each other.
# If clumping is disabled for a resource then a city's fat cross may only contain two of the resource in the following
# situations:
# 1) .......
#    ..rYY..
#    .YYYYY.
#    .YYcYY.
#    .YYYYY.
#    ..YYYr.
#    .......
#
# 2) .......
#    ..YYY..
#    .rYYYY.
#    .YYcYY.
#    .YYYYY.
#    ..YYYr.
#    .......
# where 'r' marks a resource, 'Y' marks a tile inside the city's fat cross, and 'c' marks the city tile.
# Options:
# - Standard: Attempts to approximate standard resource placement. Only luxury resources may be placed near each other.
# - Strategic: Only strategic resources may be placed near each other.
# - Health: Only health resources may be placed near each other.
# - Luxury: Only luxury resources may be placed near each other.
# - Strategic/Health: Only strategic and health resources may be placed near each other.
# - Strategic/Luxury: Only strategic and luxury resources may be placed near each other.
# - Health/Luxury: Only health and luxury resources may be placed near each other.
# - Strategic/Health/Luxury: All resources may be placed near each other.
# - Custom: The map script can be easily edited (as with other resource options) so that the user can decide which
# resources may be clumped. 'Custom' uses the user-edited clumped resources instead of one of the predefined options. If no
# editing has taken place, 'custom' uses the 'standard' option.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.6:
#
# Lands:
#
# Reduced the options 'Mountain Ranges:' and 'Changes in Altitude' down to a single option. There is no longer a range of
# options for choosing the degree to which hills and peaks congregate together, just the choice as to whether or not they
# should clump together.
#
# Terrain:
#
# Reworked terrain options. Fantasy option excluded, all terrain types use the standard terrain generator but apply
# different values to certain variables.
# Options:
# - Cool, Temperate, Warm: These determine how far from the equator ice and tundra are generated. Temperate generates both
# terrains at the normal latitudes. Cool generates both terrains closer to the equator. Warm generates both terrains
# further from the equator.
# - More desert, normal desert, less desert: How many desert tiles are generated. More desert generates roughly 50% more
# than usual. Less desert generates roughly 50% less than usual.
# - More plains, normal plains, less plains: How many plains tiles are generated. More plains generates roughly 50% more
# than usual. Less plains generates roughly 50% less than usual.
# - Fantasy: Places all land terrains randomly. Water terrain - coast or ocean - remain unchanged.
#
# Terrain Homogeneity determines how clumped each terrain type is. This option is ignored if Fantasy has been chosen.
#
# Features:
#
# An option, "Woodlands:", is now available to control the amount of forest and jungle features generated. Similar to the
# newly added terrain options, this works by giving the standard feature generator different values to work with.
# - More jungle, normal jungle, less juugle: More jungle generates more jungle than normal. Less jungle generates less
# jungle than normal.
# - More forest, normal forest, less forest: More forest generates more forests than normal. Less forest generates less
# forests than normal.
#
# An option, "Flood Plains:", is now available to give some control over placement of flood plains.
# - 'Normal flood plains placement' only places on flat, riverside desert tiles.
# - 'Extra flood plains' also places flood plains on flat, riverside grassland, plains, and tundra tiles.
#
# Resources:
#
# Setting both 'Resource Land Masses' and 'Resource Tiles' to 'Standard' will no longer deactivate all other resource
# options. Setting 'Resources' to 'Standard - ignore resource options' is now the only way to do this.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.7:
#
# Lands:
#
# Corrected an error which resulted in more than the selected percentage of ocean being generated. Previously, the amount
# of land generated was a fraction of (width - 4)*(height - 4). Now the amount of land generated is a fraction of
# width*height. This can lead to a lot more land being generated than previously. Since the amount of hills and peaks
# generated is also determined by the amount of land generated, more of these will be generated, though the percentage
# should be unchanged. The user may prefer to switch the ocean setting to a higher percentage.
#
# Reworked the code placing the initial tile of each continent. Alterations:
# - generating a small number of continents on a larger map size may take longer than previously.
# - generating a large number of continents should take a lot quicker than previously.
# - it is more likely that all requested continents are placed.
# - land masses should be more varied in shape.
#
# Added a new option, "Continent Placement:":
# - 'Normal continent placement' will place the initial tile of each continent randomly.
# - 'More rounded land masses' will generate land masses which tend to be more circular.
# - 'More snaky land masses' will generate land masses which tend to be longer and more stretched.
# Notes:
# 1) This option only affects land masses which contain 3 continents or more:
# - Selecting 2 land masses, 4 continents, and evenly distributed continents will leave the land mass shape unaffected by
# the Continent Placement options 'more rounded land masses' and 'more snaky land masses'.
# - Selecting 2 land masses, 6 continents, and evenly distributed continents will mean that the placement of the first
# tile of the last continent on each land mass will be affected by the Continent Placement options 'more rounded land
# masses' and 'more snake land masses'.
# - Selecting 2 land masses, 5 continents, and evenly distributed continents will mean that the placement of the first
# tile of the last continent on one land mass will be affected by the Continent Placement options 'more rounded land
# masses' and 'more snake land masses'.
# 2) The more continents placed on a land mass, the more noticeable the affects of the options 'more rounded land masses'
# and 'more snaky land masses'.
# - Selecting 2 land masses, 20 continents, and evenly distributed continents will make the affects of 'more rounded
# land masses' and 'more snaky land masses' more obvious than if 2 land masses, 8 continents, and evenly distributed
# continents, or 5 land masses, 20 continents, and evenly distributed continents had been selected.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.71:
#
# Lands:
#
# Since pre-2.7 LandMasses rarely produced more than 7 continents on huge maps but LandMasses 2.7 can easily generate 20
# continents on a standard size map, mountain ranges in maps generated with a large numbers of continents for every land
# mass aren't as well defined as they were previously (which, I think, removes any point in including that feature). A
# new option has been added to the Mountain Ranges option called 'ranges' (what was previously called ranges has been
# renamed 'all ranges').
# - ranges will select between 3 and 5 continents which will increase the altitude of land plots already created by other
# continents.
# - all ranges will, as before, cause all continents to increase the altitude of any land plot already created by other
# continents.
# Note:
# 1) using ranges, a mountain range produced between two continents may not be as clearly defined as one produced in
# pre-2.7 LandMasses since the extra altitude may only be generated by one of the two continents.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.72:
#
# Lands:
#
# Increased the likelihood of Indonesia-like continent-to-archipelago formations being generated under certain settings
# More specifically, when Coastlines is set to 'rougher' there is an increased chance of such formations being generated
# and when Coastlines is set to rugged there is a further increased chance of such formations being generated.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.73:
#
# Map Size:
#
# Introduced new "Min. Height" choice; 'Normal ratio of width to height'. This allows the user to select a width and an
# apporpriate height will be chosen so that the ratio of the map's width and height are similar to that generated in the
# Continent map script.
# The height chosen will be between 3/5 (0.6) and 5/8 (0.625) of the width - the minimum and maximum ratios used by
# the Continent map script (the ratio changes with different sizes).
#
# Increased the maximum value for the map width to 308 (from 240). This won't generate maps with a height of 240 using
# the 'Normal ratio of width to height' option but it allows that option to generate maps with an area equal to that of
# a 240x240 map.
#
# Lands:
#
# Reworked how continent-to-archipelago formations are generated.
#
# General Map Options:
#
# Removed Sealevel option (the one that's standard with most maps) since it wasn't doing anything.
#
# Removed standard 'Random' choice from all options. Introduced an alternative set of random options:
# - Continent Distribution: 'Random' will select from all available choices.
# - Continent Placement: 'Random' will select from all available choices.
# - Coastlines: 'Random' will select from all available choices.
# - Hills: 2 random options
#       1) 'Random' chooses between 12.5% and 34.5% of land to be hills.
#       2) 'Random highlands' chooses between 40% and 50% of land to be hills.
# - Peaks: 2 random options
#       1) 'Random' chooses between 0% and 14% of land to be peaks.
#       2) 'Random highlands' chooses between 20% and 30% of land to be peaks.
# - Mountain Ranges: 'Random' will select from all available choices.
# - Terrain: 'Random' will select from all available options except 'fantasy'.
# - Terrain Homogeneity: 'Random' will seleect from all available options.
# - Woodlands: 'Random' will select from all available options.
# - Resource Land Masses: 'Random' will select from all available options except 'custom'
# - How Many Land Masses: 'Random' will isolate resources to between 1 and (actual number of land masses - 1) land masses.
# - Strategic Quantities: 'Random' will select between 'few strategic resources', 'normal strategic resources', and 'many
# strategic resources'.
# - Health Quantities: 'Random' will select between 'few health resources', 'normal health resources', and 'many health
# resources'.
# - Luxury Quantities: 'Random' will select between 'few luxury resources', 'normal luxury resources', and 'many luxury
# resources'.
#
# Some options - usually because they are structured so that a random option is pointless or because the user is much
# more likely to have a definite idea on what they want from the option - do not recieve a random option. These options
# are:
# - Min./Max. Width
# - Min./Max. Height
# - Min./Max. Land Masses
# - Min./Max. Continents
# - Min./Max. Land Mass Buffer
# - Min./Max. Ocean
# - Flood Plains
# - Resources
# - Resource Tiles
# - Let Resources Clump
# - Print Map Details
#
# New option 'Print Map Details' so that option choices (including random selections) may be printed.
# - No: Selected options are not printed.
# - Yes: Details about the generated map are printed to C:\Program Files\Firaxis Games\Map Options. This file can be opened
# with Notepad. This file is overwritten everytime a new map is generated.
# Note that the selected option for 'Let Resources Clump:' is not listed.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.74:
#
# Lands:
#
# Fixed an error that arose when the number of continents selected for generating mountain ranges (using 'ranges') was one
# less than the number of continents generated.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.75:
#
# Lands:
#
# Fixed an error that arose when only one continent was to be generated.
#
# Resources:
#
# Fixed an error that arose when 'Resources' was set to 'standard - ignore resource options'.
#--------------------------------------------------------------------------------------------------------------------------#
# 2.90:
#
# Lands:
#
# The last land mass to be placed is more likely to have less undesirable land (tundra and snow). Previously the first
# tile of the last land mass to be placed would be placed as far from the first tile of all other land masses as possible.
# This resulted in the last land mass on maps containing more than two land masses being predominantly ice bound even if
# there was more than enough space in more temperate regions.
#
# In maps containing multiple land masses, smaller land masses should no longer be significantly smaller than the larger
# land masses. E.g. previously the third land mass in a 3 land mass map would be considerably smaller than the first two
# (which was exaggerated by the third land mass often being predominantly snow and tundra - see last listed alteration).
# Now checks have been put in place so that the map script will emphasise increasing the size of land masses that are
# particularly small and decreassing the size of land masses that are particularly large.
#
# Terrain:
#
# Removed Fantasy terrain option.
#
# Civ Placement:
#
# Included alternate civ placement strategy. Changes:
# 1) Placement code rates possible starts as follows; 2 coastal tiles > 4 coastal tiles = inland > 18 coastal tiles.
# 2) Can choose to set some land masses as being initially uninhabited.
#
# General Map Options:
#
# Expanded the number of options in the custom game set up menu. Any option which previously controlled multiple variables
# has been broken down into multiple options. E.g., the old Terrain option controlled the latitudes at which tundra and
# snow were likely to be generated, the quantity of desert, and the quantity of plains. This option has been removed and
# in its place are three options; Climate, which controls the altitude at which snow and ice are likely to be generated,
# Desert, which controls the quantity of desert, and Plains, which controls the quantity of plains.
